#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json, copy
from pathlib import Path
from tqdm import tqdm
from loguru import logger
import numpy as np
import torch
from torch.utils.data import DataLoader, TensorDataset
from torch.utils.data import RandomSampler, SequentialSampler
from ...utils.multiprocesses import barrier_leader_process, barrier_member_processes, is_multi_processes
from ...utils import seg_generator
from ..ner_utils import InputExample


class InputFeature(object):
    """A single set of features of data."""
    def __init__(self, input_ids, input_mask, input_len, segment_ids,
                 start_ids, end_ids, subjects):
        self.input_ids = input_ids
        self.input_mask = input_mask
        self.segment_ids = segment_ids
        self.start_ids = start_ids
        self.input_len = input_len
        self.end_ids = end_ids
        self.subjects = subjects

    def __repr__(self):
        return str(self.to_json_string())

    def __eq__(self, other):
        return self.input_ids == other.input_ids and \
            self.input_mask == other.input_mask and \
            self.segment_ids == other.segment_ids and \
            self.start_ids == other.start_ids and \
            self.end_ids == other.end_ids and \
            self.subjects == other.subjects and \
            self.input_len == other.input_len

    def to_dict(self):
        """Serializes this instance to a Python dictionary."""
        output = copy.deepcopy(self.__dict__)
        return output

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        return json.dumps(self.to_dict(), indent=2, sort_keys=True) + "\n"


#  def load_examples(args,
#                    data_generator,
#                    examples_file,
#                    seg_len=0,
#                    seg_backoff=0):
#
#      examples = []
#
#      for guid, text_a, text_b, labels in data_generator(
#              args, examples_file, seg_len=seg_len, seg_backoff=seg_backoff):
#          assert text_a is not None
#          #  for (seg_text_a, seg_text_b) in seg_generator((text_a, text_b),
#          #                                                seg_len, seg_backoff):
#          #      seg_words_a = [w for w in seg_text_a]
#          #
#          #      examples.append(
#          #          InputExample(guid=guid, text_a=seg_words_a, labels=labels))
#          examples.append(InputExample(guid=guid, text_a=text_a, labels=labels))
#      logger.info(f"Loaded {len(examples)} examples.")
#
#      return examples


def encode_examples(examples, label2id, tokenizer, max_seq_length):

    #  texts = [''.join(e.text_a) for e in examples]
    texts = [e.text_a[:max_seq_length - 2] for e in examples]

    #  outputs = tokenizer.batch_encode_plus(texts,
    #                                        max_length=max_seq_length,
    #                                        add_special_tokens=True,
    #                                        return_tensors='pt',
    #                                        pad_to_max_length=True,
    #                                        return_attention_mask=True,
    #                                        return_token_type_ids=True)
    #  all_input_ids, all_attention_mask, all_token_type_ids = outputs[
    #      'input_ids'], outputs['attention_mask'], outputs['token_type_ids']

    all_tokens = [[tokenizer.cls_token] + tokenizer.tokenize(text) +
                  [tokenizer.sep_token]
                  for text in tqdm(texts, desc="Tokenize")]
    all_input_lens = [len(tokens) for tokens in all_tokens]
    all_input_ids = [
        tokenizer.convert_tokens_to_ids(tokens) for tokens in all_tokens
    ]
    all_attention_mask = [[1] * len(input_ids) for input_ids in all_input_ids]
    all_token_type_ids = [[0] * len(input_ids) for input_ids in all_input_ids]

    all_padding_lens = [max_seq_length - n for n in all_input_lens]
    for i, (input_ids, attention_mask, token_type_ids,
            padding_length) in enumerate(
                zip(all_input_ids, all_attention_mask, all_token_type_ids,
                    all_padding_lens)):
        all_input_ids[i] = input_ids + [0] * padding_length
        all_attention_mask[i] = attention_mask + [0] * padding_length
        all_token_type_ids[i] = token_type_ids + [0] * padding_length

    def encode_subjects(subjects, num_tokens):
        #  logger.warning(f"num_tokens: {num_tokens}")
        start_ids = [label2id['[unused1]']] * num_tokens
        end_ids = [label2id['[unused1]']] * num_tokens
        subjects_id = []
        if subjects:
            for subject in subjects:
                #  logger.debug(f"subject: {subject}")
                label = subject[0]
                start = subject[1]
                end = subject[2]
                start_ids[start + 1] = label2id[label]
                end_ids[end + 1] = label2id[label]
                subjects_id.append((label2id[label], start, end))
        return subjects_id, start_ids, end_ids

    all_subjects_ids = []
    all_start_ids = []
    all_end_ids = []
    all_subjects = [e.labels for e in examples]
    for input_ids, attention_mask, token_type_ids, subjects in zip(
            all_input_ids, all_attention_mask, all_token_type_ids,
            all_subjects):
        num_tokens = len(input_ids)
        subjects_id, start_ids, end_ids = encode_subjects(subjects, num_tokens)

        assert len(input_ids) == max_seq_length
        assert len(attention_mask) == max_seq_length
        assert len(token_type_ids) == max_seq_length
        assert len(start_ids) == max_seq_length
        assert len(end_ids) == max_seq_length

        all_subjects_ids.append(subjects_id)
        all_start_ids.append(start_ids)
        all_end_ids.append(end_ids)

    #  all_start_ids = torch.tensor(all_start_ids, dtype=torch.long)
    #  all_end_ids = torch.tensor(all_end_ids, dtype=torch.long)
    #  all_input_lens = torch.tensor(all_input_lens, dtype=torch.long)

    #  all_input_lens = torch.tensor(
    #      [len(input_ids) for input_ids in all_input_ids], dtype=torch.long)

    logger.debug(f"all_input_ids.shape: {np.array(all_input_ids).shape}")
    logger.debug(
        f"all_attention_mask.shape: {np.array(all_attention_mask).shape}")
    logger.debug(
        f"all_token_type_ids.shape: {np.array(all_token_type_ids).shape}")
    logger.debug(f"all_start_ids.shape: {np.array(all_start_ids).shape}")
    logger.debug(f"all_end_ids.shape: {np.array(all_end_ids).shape}")
    logger.debug(f"all_subjects_ids.shape: {np.array(all_subjects_ids).shape}")
    logger.debug(f"all_input_lens.shape: {np.array(all_input_lens).shape}")
    assert np.array(all_input_ids).shape[1] == max_seq_length
    assert np.array(all_attention_mask).shape[1] == max_seq_length
    assert np.array(all_token_type_ids).shape[1] == max_seq_length

    all_features = [
        InputFeature(input_ids=input_ids,
                     input_mask=attention_mask,
                     segment_ids=token_type_ids,
                     start_ids=start_ids,
                     end_ids=end_ids,
                     subjects=subjects_ids,
                     input_len=input_len) for input_ids, attention_mask,
        token_type_ids, start_ids, end_ids, subjects_ids, input_len in zip(
            all_input_ids, all_attention_mask, all_token_type_ids,
            all_start_ids, all_end_ids, all_subjects_ids, all_input_lens)
    ]
    return (all_features, {
        'input_ids':
        torch.tensor(all_input_ids, dtype=torch.long),
        'attention_mask':
        torch.tensor(all_attention_mask, dtype=torch.long),
        'token_type_ids':
        torch.tensor(all_token_type_ids, dtype=torch.long),
        'start_ids':
        torch.tensor(all_start_ids, dtype=torch.long),
        'end_ids':
        torch.tensor(all_end_ids, dtype=torch.long),
        "subjects_ids":
        all_subjects_ids,
        "input_lens":
        torch.tensor(all_input_lens, dtype=torch.long),
    })


def examples_to_dataset(examples, label2id, tokenizer, max_seq_length):
    all_features, outputs = encode_examples(examples, label2id, tokenizer,
                                            max_seq_length)

    #  dataset_old, features_old = examples_to_dataset_old(
    #      examples, label2id, tokenizer, max_seq_length)
    #
    #  for x, x1 in zip(all_features, features_old):
    #      if x != x1:
    #          logger.info(f"input_ids: {len(x.input_ids)}, {x.input_ids}")
    #          logger.info(f"input_mask: {len(x.input_mask)}, {x.input_mask}")
    #          logger.info(f"segment_ids: {len(x.segment_ids)}, {x.segment_ids}")
    #          logger.info(f"start_ids: {len(x.start_ids)}, {x.start_ids}")
    #          logger.info(f"end_ids: {len(x.end_ids)}, {x.end_ids}")
    #          logger.info(f"subjects: {x.subjects}")
    #          logger.info(f"input_len: {x.input_len}")
    #
    #          if x.input_ids != x1.input_ids:
    #              logger.warning(
    #                  f"input_ids: {len(x1.input_ids)}, {x1.input_ids}")
    #          if x.input_mask != x1.input_mask:
    #              logger.warning(
    #                  f"input_mask: {len(x1.input_mask)}, {x1.input_mask}")
    #          if x.segment_ids != x1.segment_ids:
    #              logger.warning(
    #                  f"segment_ids: {len(x1.segment_ids)}, {x1.segment_ids}")
    #          if x.start_ids != x1.start_ids:
    #              logger.warning(
    #                  f"start_ids: {len(x1.start_ids)}, {x1.start_ids}")
    #          if x.end_ids != x1.end_ids:
    #              logger.warning(f"end_ids: {len(x1.end_ids)}, {x1.end_ids}")
    #          if x.subjects != x1.subjects:
    #              logger.warning(f"subjects: {x1.subjects}")
    #          if x.input_len != x1.input_len:
    #              logger.warning(f"input_len: {x1.input_len}")

    dataset = TensorDataset(
        outputs['input_ids'],
        outputs['attention_mask'],
        outputs['token_type_ids'],
        outputs['start_ids'],
        outputs['end_ids'],
        outputs['input_lens'],
    )

    assert len(dataset) == len(all_features)

    return dataset, all_features


def convert_examples_to_features(
    examples,
    label2id,
    tokenizer,
    max_seq_length,
    cls_token_at_end=False,
    cls_token="[CLS]",
    cls_token_segment_id=1,
    sep_token="[SEP]",
    pad_on_left=False,
    pad_token=0,
    pad_token_segment_id=0,
    sequence_a_segment_id=0,
    mask_padding_with_zero=True,
):
    """ Loads a data file into a list of `InputBatch`s
        `cls_token_at_end` define the location of the CLS token:
            - False (Default, BERT/XLM pattern): [CLS] + A + [SEP] + B + [SEP]
            - True (XLNet/GPT pattern): A + [SEP] + B + [SEP] + [CLS]
        `cls_token_segment_id` define the segment id associated to the CLS token (0 for BERT, 2 for XLNet)
    """
    features = []
    for ex_index, e in enumerate(tqdm(examples, desc="Convert examples")):
        textlist = e.text_a
        subjects = e.labels

        #  logger.warning(f"textlist: {textlist}")
        tokens = tokenizer.tokenize(textlist)
        #  tokens = tokenizer.tokenize(''.join(textlist))[1:-1]

        #  logger.debug(f"tokens: {tokens}")
        #  logger.debug(f"textlist: {textlist}")
        #  logger.warning(
        #      f"tokens: {len(tokens)} - {tokens[:10]} - {tokens[-10:]}")
        #  logger.warning(
        #      f"textlist: {len(textlist)} - {textlist[:10]} - {textlist[-10:]}")
        #  assert len(tokens) == len(textlist)

        #  logger.debug(f"len(tokens): {len(tokens)}")
        start_ids = [0] * len(tokens)
        end_ids = [0] * len(tokens)
        subjects_id = []
        if subjects:
            for subject in subjects:
                #  logger.debug(f"subject: {subject}")
                label = subject[0]
                start = subject[1]
                end = subject[2]
                #  logger.warning(
                #      f"len(tokens): {len(tokens)}, subject: {subject}")
                start_ids[start] = label2id[label]
                end_ids[end] = label2id[label]
                subjects_id.append((label2id[label], start, end))
        # Account for [CLS] and [SEP] with "- 2".
        special_tokens_count = 2
        if len(tokens) > max_seq_length - special_tokens_count:
            tokens = tokens[:(max_seq_length - special_tokens_count)]
            start_ids = start_ids[:(max_seq_length - special_tokens_count)]
            end_ids = end_ids[:(max_seq_length - special_tokens_count)]

        # The convention in BERT is:
        # (a) For sequence pairs:
        #  tokens:   [CLS] is this jack ##son ##ville ? [SEP] no it is not . [SEP]
        #  type_ids:   0   0  0    0    0     0       0   0   1  1  1  1   1   1
        # (b) For single sequences:
        #  tokens:   [CLS] the dog is hairy . [SEP]
        #  type_ids:   0   0   0   0  0     0   0
        #
        # Where "type_ids" are used to indicate whether this is the first
        # sequence or the second sequence. The embedding vectors for `type=0` and
        # `type=1` were learned during pre-training and are added to the wordpiece
        # embedding vector (and position vector). This is not *strictly* necessary
        # since the [SEP] token unambiguously separates the sequences, but it makes
        # it easier for the model to learn the concept of sequences.
        #
        # For classification tasks, the first vector (corresponding to [CLS]) is
        # used as as the "sentence vector". Note that this only makes sense because
        # the entire model is fine-tuned.
        tokens += [sep_token]
        start_ids += [0]
        end_ids += [0]
        segment_ids = [sequence_a_segment_id] * len(tokens)
        if cls_token_at_end:
            tokens += [cls_token]
            start_ids += [0]
            end_ids += [0]
            segment_ids += [cls_token_segment_id]
        else:
            tokens = [cls_token] + tokens
            start_ids = [0] + start_ids
            end_ids = [0] + end_ids
            segment_ids = [cls_token_segment_id] + segment_ids

        input_ids = tokenizer.convert_tokens_to_ids(tokens)

        # The mask has 1 for real tokens and 0 for padding tokens. Only real
        # tokens are attended to.
        input_mask = [1 if mask_padding_with_zero else 0] * len(input_ids)
        input_len = len(input_ids)
        # Zero-pad up to the sequence length.
        padding_length = max_seq_length - len(input_ids)
        if pad_on_left:
            input_ids = ([pad_token] * padding_length) + input_ids
            input_mask = ([0 if mask_padding_with_zero else 1] *
                          padding_length) + input_mask
            segment_ids = ([pad_token_segment_id] *
                           padding_length) + segment_ids
            start_ids = ([0] * padding_length) + start_ids
            end_ids = ([0] * padding_length) + end_ids
        else:
            input_ids += [pad_token] * padding_length
            input_mask += [0 if mask_padding_with_zero else 1] * padding_length
            segment_ids += [pad_token_segment_id] * padding_length
            start_ids += ([0] * padding_length)
            end_ids += ([0] * padding_length)

        assert len(input_ids) == max_seq_length
        assert len(input_mask) == max_seq_length
        assert len(segment_ids) == max_seq_length
        assert len(start_ids) == max_seq_length
        assert len(end_ids) == max_seq_length

        if ex_index < 3:
            logger.debug("*** Example ***")
            logger.debug(f"guid: {e.guid}")
            logger.debug(f"tokens: {' '.join([str(x) for x in tokens])}")
            logger.debug(f"input_ids: {' '.join([str(x) for x in input_ids])}")
            logger.debug(
                f"input_mask: {' '.join([str(x) for x in input_mask])}")
            logger.debug(
                f"segment_ids: {' '.join([str(x) for x in segment_ids])}")
            logger.info(f"start_ids: {' '.join([str(x) for x in start_ids])}")
            logger.info(f"end_ids: {' '.join([str(x) for x in end_ids])}")

            #  logger.info("*** Example ***")
            #  logger.info(f"guid: %s", e.guid)
            #  logger.info("tokens: %s", " ".join([str(x) for x in tokens]))
            #  logger.info("input_ids: %s", " ".join([str(x) for x in input_ids]))
            #  logger.info("input_mask: %s",
            #              " ".join([str(x) for x in input_mask]))
            #  logger.info("segment_ids: %s",
            #              " ".join([str(x) for x in segment_ids]))
            #  logger.info("start_ids: %s" % " ".join([str(x)
            #                                          for x in start_ids]))
            #  logger.info("end_ids: %s" % " ".join([str(x) for x in end_ids]))

        features.append(
            InputFeature(input_ids=input_ids,
                         input_mask=input_mask,
                         segment_ids=segment_ids,
                         start_ids=start_ids,
                         end_ids=end_ids,
                         subjects=subjects_id,
                         input_len=input_len))
    return features


def features_to_dataset(features):
    all_input_ids = torch.tensor([f.input_ids for f in features],
                                 dtype=torch.long)
    all_input_mask = torch.tensor([f.input_mask for f in features],
                                  dtype=torch.long)
    all_segment_ids = torch.tensor([f.segment_ids for f in features],
                                   dtype=torch.long)
    all_start_ids = torch.tensor([f.start_ids for f in features],
                                 dtype=torch.long)
    all_end_ids = torch.tensor([f.end_ids for f in features], dtype=torch.long)
    all_input_lens = torch.tensor([f.input_len for f in features],
                                  dtype=torch.long)
    #  all_subjects_ids = torch.tensor([f.subjects for f in features],
    #                                  dtype=torch.long)
    dataset = TensorDataset(all_input_ids, all_input_mask, all_segment_ids,
                            all_start_ids, all_end_ids, all_input_lens)
    return dataset


def examples_to_dataset_with_cache(args,
                                   examples,
                                   tokenizer,
                                   max_seq_length,
                                   evaluate=False,
                                   alias=""):

    # Make sure only the first process in distributed training process the dataset,
    # and the others will use the cache
    if not evaluate:
        barrier_member_processes(args)

    # Load data features from cache or dataset file
    cached_features_file = Path(
        args.data_dir
    ) / f"cached_ner_{args.dataset_name}{'_' + alias if alias else ''}"
    if args.cache_features and cached_features_file.exists(
    ) and not args.overwrite_cache:
        logger.info(
            f"Loading features from cached file {cached_features_file}")
        features = torch.load(cached_features_file)
    else:
        logger.info(f"Creating features from dataset file at {args.data_dir}")

        features = convert_examples_to_features(
            examples,
            args.label2id,
            tokenizer,
            max_seq_length=max_seq_length,
            cls_token_at_end=bool(args.model_type in ["xlnet"]),
            pad_on_left=bool(args.model_type in ['xlnet']),
            cls_token=tokenizer.cls_token,
            cls_token_segment_id=2 if args.model_type in ["xlnet"] else 0,
            sep_token=tokenizer.sep_token,
            # pad on the left for xlnet
            pad_token=tokenizer.convert_tokens_to_ids([tokenizer.pad_token
                                                       ])[0],
            pad_token_segment_id=4 if args.model_type in ['xlnet'] else 0,
        )
        if args.cache_features and is_master_process(args):
            logger.info(
                f"Saving features into cached file {cached_features_file}")
            torch.save(features, cached_features_file)

    # Make sure only the first process in distributed training process the dataset,
    # and the others will use the cache
    if not evaluate:
        barrier_leader_process(args)

    dataset = features_to_dataset(features)

    return dataset, features


def examples_to_dataset_old(examples, label2id, tokenizer, max_seq_length):

    features = convert_examples_to_features(
        examples,
        label2id,
        tokenizer,
        max_seq_length=max_seq_length,
        cls_token_at_end=False,
        pad_on_left=False,
        cls_token=tokenizer.cls_token,
        cls_token_segment_id=0,
        sep_token=tokenizer.sep_token,
        pad_token=tokenizer.convert_tokens_to_ids([tokenizer.pad_token])[0],
        pad_token_segment_id=0)

    dataset = features_to_dataset(features)

    return dataset, features



# -*- coding: utf-8 -*-
import os, re
import numpy as np
from loguru import logger

#  import tensorflow as tf
from keras.models import Model
from keras.layers import Dense, Lambda, Input
import keras.backend as K
from keras.optimizers import Adam, RMSprop
from adamw import AdamW
from keras.utils import multi_gpu_model

#  os.environ.setdefault('TF_KERAS', '1')
print(f"TF_KERAS={os.environ.get('TF_KERAS', '0')}")
from keras_bert import load_trained_model_from_checkpoint, Tokenizer

from train_config import Config

# for keras_bert.load_trained_model_from_checkpoint()
#  bert_seq_len = 512
#  extract_max_length = Config.max_seq_length + 2  # bert_seq_len  # - 2  #180  #1024

debug_extract_labels = Config.debug_extract_labels


#%% ------------------------------------------------------------
# Model
class BertTokenizer(Tokenizer):
    def __init__(self, bert_vocab_file: str):
        tokens = {}
        for line in open(bert_vocab_file, 'r').readlines():
            token = line.strip()
            tokens[token] = len(tokens)
        Tokenizer.__init__(self, tokens)

    def _tokenize(self, text):
        return [
            x if x in self._token_dict else
            # space类用未经训练的[unused1]表示，
            # 未在字典中的字符用[UNK]表示。
            ('[unused1]' if self._is_space(x) else '[UNK]') for x in text
        ]


def load_bert_pretrained_model(max_seq_length):
    extract_max_length = max_seq_length + 2
    bert_pretrained_model = load_trained_model_from_checkpoint(
        Config.config_path, Config.checkpoint_path, seq_len=extract_max_length)
    if Config.gpus > 1:
        bert_pretrained_model = multi_gpu_model(bert_pretrained_model,
                                                gpus=Config.gpus)
    for layer in bert_pretrained_model.layers:
        layer.trainable = True

    return bert_pretrained_model


def find_pos_pairs(ps1, ps2, s_threshold=0.01, e_threshold=0.01, top_k=0):
    assert len(ps1) == len(ps2)
    s = [(prob, i) for i, prob in enumerate(ps1)]
    e = [(prob, i) for i, prob in enumerate(ps2)]
    s = sorted(s,
               key=lambda x: f"{x[0]:.4f},{1 /(x[1] + 1):.6f}",
               reverse=True)
    e = sorted(e,
               key=lambda x: f"{x[0]:.4f},{1 /(x[1] + 1):.6f}",
               reverse=True)
    #  logger.debug(f"s: {s}")
    #  logger.debug(f"e: {e}")
    s = [x for x in s if x[0] >= s_threshold]
    e = [x for x in e if x[0] >= e_threshold]
    if top_k > 0:
        s = s[:top_k]
        e = e[:top_k]

    #  logger.debug(f"filtered s: {s}")
    #  logger.debug(f"filtered e: {e}")

    return s, e


def find_entities_from_indices(indices, min_label_len):
    dangling_entity = None
    s_idx = -1
    for s_idx, (_, _, (tag, idx, prob, _)) in enumerate(indices):
        if tag == 's':
            break
    pairs = []
    if s_idx >= 0:
        text_in_0, c_in_0, (tag0, idx0, prob0, text_offset) = indices[s_idx]
        idx0 += text_offset
        indices = indices[s_idx + 1:]
        prob1 = 0.0
        for text_in, c_in, (tag, idx, prob, text_offset) in indices:
            idx += text_offset

            if tag == tag0:
                if tag0 == 's':
                    if prob > prob0:
                        idx0 = idx
                        prob0 = prob
                        text_in_0 = text_in
                        c_in_0 = c_in
                else:
                    if prob > prob1:
                        idx1 = idx
                        prob1 = prob
                        pairs[-1] = (pairs[-1][0], (idx1, prob1), text_in,
                                     c_in)
            else:
                if tag0 == 's':
                    # 's' -> 'e'
                    idx1 = idx
                    prob1 = prob
                    pairs.append(((idx0, prob0), (idx1, prob1), text_in, c_in))
                else:
                    # 'e' -> 's'
                    idx0 = idx
                    prob0 = prob
                    text_in_0 = text_in
                    c_in_0 = c_in
                tag0 = tag
                idx0 = idx
                text_in_0 = text_in
                c_in_0 = c_in
        if tag0 == 's':
            #  dangling_entity = (idx0 if idx0 == 0 else idx0 - 1, prob0,
            dangling_entity = (idx0, prob0, text_in_0, c_in_0)
    #  pairs = [((idx0 if idx0 == 0 else idx0 - 1, prob0), (idx1, prob1), text_in,

    pairs = [((idx0, prob0), (idx1, prob1), text_in, c_in)
             for (idx0, prob0), (idx1, prob1), text_in, c_in in pairs
             if idx1 - idx0 >= min_label_len]

    entities = [(idx0, idx1, prob0, prob1, text_in, c_in)
                for (idx0, prob0), (idx1, prob1), text_in, c_in in pairs]

    return entities, dangling_entity


class BertModel():
    def __init__(self, config_path, checkpoint_path, dict_path, learning_rate,
                 max_seq_length):

        self.tokenizer = BertTokenizer(dict_path)
        self.learning_rate = learning_rate
        self.x1_in = Input(shape=(None, ))  # 待识别句子输入
        self.x2_in = Input(shape=(None, ))  # 待识别句子输入
        self.s1_in = Input(shape=(None, ))  # 实体左边界（标签）
        self.s2_in = Input(shape=(None, ))  # 实体右边界（标签）

        #  bert_pretrained_model = load_trained_model_from_checkpoint(
        #      config_path, checkpoint_path, seq_len=extract_max_length)
        #  if Config.gpus > 1:
        #      bert_pretrained_model = multi_gpu_model(bert_pretrained_model,
        #                                              gpus=Config.gpus)
        #
        #  for layer in bert_pretrained_model.layers:
        #      layer.trainable = True
        bert_pretrained_model = load_bert_pretrained_model(max_seq_length)

        self.bert_pretrained_model = bert_pretrained_model  #to_tpu_model(bert_pretrained_model)

    def build_model(self):
        x1, x2, s1, s2 = self.x1_in, self.x2_in, self.s1_in, self.s2_in
        self.x_mask = Lambda(
            lambda x: K.cast(K.greater(K.expand_dims(x, 2), 0), 'float32'))(x1)

        x = self.bert_pretrained_model([x1, x2])

        # BiGRU + DNN
        #  # from https://github.com/hecongqing/CCKS2019EventEntityExtraction_Rank5/blob/master/src/SEBERT_model.py
        #  l = Lambda(lambda t: t[:, -1])(x)
        #  x = Add()([x, l])
        #  x = Dropout(0.1)(x)
        #  x = Lambda(lambda x: x[0] * x[1])([x, x_mask])
        #
        #  x = SpatialDropout1D(0.1)(x)
        #  x = Bidirectional(CuDNNGRU(200, return_sequences=True))(x)
        #  x = Lambda(lambda x: x[0] * x[1])([x, x_mask])
        #  x = Bidirectional(CuDNNGRU(200, return_sequences=True))(x)
        #  x = Lambda(lambda x: x[0] * x[1])([x, x_mask])
        #
        #  x = Dense(1024, use_bias=False, activation='tanh')(x)
        #  x = Dropout(0.2)(x)
        #  x = Dense(64, use_bias=False, activation='tanh')(x)
        #  x = Dropout(0.2)(x)
        #  x = Dense(8, use_bias=False, activation='tanh')(x)

        ps1 = Dense(1, use_bias=False)(x)
        ps1 = Lambda(lambda x: x[0][..., 0] - (1 - x[1][..., 0]) * 1e10)(
            [ps1, self.x_mask])
        ps2 = Dense(1, use_bias=False)(x)
        ps2 = Lambda(lambda x: x[0][..., 0] - (1 - x[1][..., 0]) * 1e10)(
            [ps2, self.x_mask])

        self.predict_model = Model([self.x1_in, self.x2_in], [ps1, ps2])

        train_model = Model([self.x1_in, self.x2_in, self.s1_in, self.s2_in],
                            [ps1, ps2])
        if Config.gpus > 1:
            train_model = multi_gpu_model(train_model, gpus=Config.gpus)

        def get_loss(y_true, y_pred, with_weights=False):
            weights = 0.0
            if with_weights:
                # 根据标签位置距离，计算分类权重。
                i_true = K.argmax(y_true, axis=1)
                i_pred = K.argmax(y_pred, axis=1)
                distance = K.abs(i_true - i_pred)
                weights = K.cast(distance, dtype='float32')
                #  length = K.int_shape(y_true)[1] - 1
                #  weights = K.cast(distance / length, dtype='float32')

            losses = (1.0 + weights) * K.categorical_crossentropy(
                y_true, y_pred, from_logits=True)
            #  losses = (
            #      (1.0 + weights) *
            #      K.categorical_crossentropy(y_true, y_pred, from_logits=True))
            loss = K.mean(losses)
            return loss

        #  loss1 = K.mean(
        #      K.categorical_crossentropy(self.s1_in, ps1, from_logits=True))
        #  ps2 -= (1 - K.cumsum(s1, 1)) * 1e10
        #  loss2 = K.mean(
        #      K.categorical_crossentropy(self.s2_in, ps2, from_logits=True))
        #  self.loss = loss1 + loss2
        loss1 = get_loss(self.s1_in, ps1, with_weights=True)
        ps2 -= (1 - K.cumsum(s1, 1)) * 1e10
        loss2 = get_loss(self.s2_in, ps2, with_weights=True)
        self.loss = loss1 + loss2

        train_model.add_loss(self.loss)

        if 'COLAB_TPU_ADDR' in os.environ:
            train_model.compile(
                #optimizer=tf.train.RMSPropOptimizer(self.learning_rate))
                optimizer=RMSprop())
        else:
            #  from accum_optimizer import AccumOptimizer
            #  train_model.compile(optimizer=AccumOptimizer(
            #      Adam(self.learning_rate), steps_per_update))
            train_model.compile(optimizer=AdamW(self.learning_rate))
        train_model.summary()

        self.train_model = train_model

    def determine_entity_positions(self,
                                   text_in,
                                   c_in,
                                   text_offset,
                                   top_k=20,
                                   s_threshold=0.01,
                                   e_threshold=0.01,
                                   seg_len=510,
                                   seg_backoff=400):
        def softmax(x):
            x = x - np.max(x)
            x = np.exp(x)
            return x / np.sum(x)

        text_prefix = f"___{c_in}___"
        text_in = f"{text_prefix}{text_in}"
        #  text_in = u'___%s___%s' % (c_in, text_in)
        text_in = text_in[:seg_len]
        #  text_in += " " * (seg_len- len(text_in))

        _tokens = self.tokenizer.tokenize(text_in)
        num_tokens = len(_tokens)
        _x1, _x2 = self.tokenizer.encode(first=text_in)
        _x1, _x2 = np.array([_x1]), np.array([_x2])
        #  print(f"_x1.shape: {_x1.shape} _x2.shape: {_x2.shape}")
        #  print(f"{len(text_in)} : {text_in[:30]}")
        _ps1, _ps2 = self.predict_model.predict([_x1, _x2])
        #  print(f"_ps1: {_ps1}")
        #  print(f"_ps2: {_ps2}")
        _ps1, _ps2 = softmax(_ps1[0]), softmax(_ps2[0])
        #  print(f"-------- after softmax --------")
        #  print(f"_ps1: {_ps1}")
        #  print(f"_ps2: {_ps2}")

        s_indices, e_indices = find_pos_pairs(_ps1,
                                              _ps2,
                                              s_threshold=s_threshold,
                                              e_threshold=e_threshold,
                                              top_k=top_k)
        #  print("s_indices:", s_indices)
        #  print("e_indices:", e_indices)
        indices = [('s', idx - len(text_prefix) - 1, prob, text_offset)
                   for prob, idx in s_indices
                   ] + [('e', idx - len(text_prefix), prob, text_offset)
                        for prob, idx in e_indices]
        indices = sorted(indices, key=lambda x: x[1])
        if debug_extract_labels:
            logger.debug(f"indices: {indices}")

        return indices

    #  def extract_entity(self, text_in, c_in, additional_chars):
    #      def softmax(x):
    #          x = x - np.max(x)
    #          x = np.exp(x)
    #          return x / np.sum(x)
    #
    #      text_in = u'___%s___%s' % (c_in, text_in)
    #      text_in = text_in[:Config.extract_max_length]
    #      #  text_in += " " * (Config.extract_max_length - len(text_in))
    #
    #      _tokens = self.tokenizer.tokenize(text_in)
    #      num_tokens = len(_tokens)
    #      _x1, _x2 = self.tokenizer.encode(first=text_in)
    #      _x1, _x2 = np.array([_x1]), np.array([_x2])
    #      #  print(f"_x1.shape: {_x1.shape} _x2.shape: {_x2.shape}")
    #      #  print(f"{len(text_in)} : {text_in[:30]}")
    #      _ps1, _ps2 = self.predict_model.predict([_x1, _x2])
    #      #  print(f"_ps1: {_ps1}")
    #      #  print(f"_ps2: {_ps2}")
    #      _ps1, _ps2 = softmax(_ps1[0]), softmax(_ps2[0])
    #      #  print(f"-------- after softmax --------")
    #      #  print(f"_ps1: {_ps1}")
    #      #  print(f"_ps2: {_ps2}")
    #
    #      s_indices, e_indices = find_pos_pairs(_ps1,
    #                                            _ps2,
    #                                            s_threshold=0.1,
    #                                            e_threshold=0.01,
    #                                            top_k=20)
    #
    #      #  print("s_indices:", s_indices)
    #      #  print("e_indices:", e_indices)
    #      indices = [('s', idx, prob)
    #                 for prob, idx in s_indices] + [('e', idx, prob)
    #                                                for prob, idx in e_indices]
    #      indices = sorted(indices, key=lambda x: x[1])
    #      #  print("indices:", indices)
    #
    #      s_idx = -1
    #      for s_idx, (tag, idx, prob) in enumerate(indices):
    #          if tag == 's':
    #              break
    #      pairs = []
    #      if s_idx >= 0:
    #          tag0, idx0, prob0 = indices[s_idx]
    #          indices = indices[s_idx + 1:]
    #          for tag, idx, prob in indices:
    #              if tag == tag0:
    #                  if prob > prob0:
    #                      if tag0 == 's':
    #                          idx0 = idx
    #                          prob0 = prob
    #                      else:
    #                          idx1 = idx
    #                          prob1 = prob
    #                          pairs[-1] = (pairs[-1][0], (idx1, prob1))
    #              else:
    #                  if tag0 == 's':
    #                      # 's' -> 'e'
    #                      idx1 = idx
    #                      prob1 = prob
    #                      pairs.append(((idx0, prob0), (idx1, prob1)))
    #                  else:
    #                      # 'e' -> 's'
    #                      idx0 = idx
    #                      prob0 = prob
    #                  tag0 = tag
    #                  idx0 = idx
    #                  prob0 = prob
    #
    #      pairs = [((idx0, prob0), (idx1, prob1))
    #               for (idx0, prob0), (idx1, prob1) in pairs if idx1 - idx0 > 3]
    #      #  print("pairs:", pairs)
    #      pairs = [(idx0, idx1) for (idx0, _), (idx1, _) in pairs]
    #      #  for idx0, idx1 in pairs:
    #      #      print(f"-------------------- ({idx0},{idx1}) --------------------")
    #      #      print(text_in[idx0:idx1])
    #      #      print("")
    #
    #      for i, _t in enumerate(_tokens):
    #          if len(_t) == 1 and re.findall(u'[^\u4e00-\u9fa5a-zA-Z0-9\*]',
    #                                         _t) and _t not in additional_chars:
    #              _ps1[i] -= 10
    #      start = _ps1.argmax()
    #      for end in range(start, num_tokens):
    #          _t = _tokens[end]
    #          if len(_t) == 1 and re.findall(u'[^\u4e00-\u9fa5a-zA-Z0-9\*]',
    #                                         _t) and _t not in additional_chars:
    #              break
    #      end = _ps2[start:end + 1].argmax() + start
    #      #  print(f"start: {start}, end: {end}")
    #      a = text_in[start - 1:end]
    #
    #      k1, k2 = np.zeros(num_tokens), np.zeros(num_tokens)
    #      if start >= 0:
    #          k1[start] = 1
    #          k2[end] = 1
    #
    #      return a, (k1.tolist(), k2.tolist())

    #  def extract_entities(self,
    #                       text_in,
    #                       c_in,
    #                       top_k=20,
    #                       s_threshold=0.01,
    #                       e_threshold=0.01,
    #                       min_label_len=8):
    #      def softmax(x):
    #          x = x - np.max(x)
    #          x = np.exp(x)
    #          return x / np.sum(x)
    #
    #      text_in = u'___%s___%s' % (c_in, text_in)
    #      text_in = text_in[:Config.extract_max_length]
    #      #  text_in += " " * (Config.extract_max_length - len(text_in))
    #
    #      dangling_entity = None
    #
    #      _tokens = self.tokenizer.tokenize(text_in)
    #      num_tokens = len(_tokens)
    #      _x1, _x2 = self.tokenizer.encode(first=text_in)
    #      _x1, _x2 = np.array([_x1]), np.array([_x2])
    #      #  print(f"_x1.shape: {_x1.shape} _x2.shape: {_x2.shape}")
    #      #  print(f"{len(text_in)} : {text_in[:30]}")
    #      _ps1, _ps2 = self.predict_model.predict([_x1, _x2])
    #      #  print(f"_ps1: {_ps1}")
    #      #  print(f"_ps2: {_ps2}")
    #      _ps1, _ps2 = softmax(_ps1[0]), softmax(_ps2[0])
    #      #  print(f"-------- after softmax --------")
    #      #  print(f"_ps1: {_ps1}")
    #      #  print(f"_ps2: {_ps2}")
    #
    #      s_indices, e_indices = find_pos_pairs(_ps1,
    #                                            _ps2,
    #                                            s_threshold=s_threshold,
    #                                            e_threshold=e_threshold,
    #                                            top_k=top_k)
    #      #  print("s_indices:", s_indices)
    #      #  print("e_indices:", e_indices)
    #      indices = [('s', idx, prob)
    #                 for prob, idx in s_indices] + [('e', idx, prob)
    #                                                for prob, idx in e_indices]
    #      indices = sorted(indices, key=lambda x: x[1])
    #      #  print("indices:", indices)
    #
    #      s_idx = -1
    #      for s_idx, (tag, idx, prob) in enumerate(indices):
    #          if tag == 's':
    #              break
    #      pairs = []
    #      if s_idx >= 0:
    #          tag0, idx0, prob0 = indices[s_idx]
    #          indices = indices[s_idx + 1:]
    #          for tag, idx, prob in indices:
    #              if tag == tag0:
    #                  if prob > prob0:
    #                      if tag0 == 's':
    #                          idx0 = idx
    #                          prob0 = prob
    #                      else:
    #                          idx1 = idx
    #                          prob1 = prob
    #                          pairs[-1] = (pairs[-1][0], (idx1, prob1))
    #              else:
    #                  if tag0 == 's':
    #                      # 's' -> 'e'
    #                      idx1 = idx
    #                      prob1 = prob
    #                      pairs.append(((idx0, prob0), (idx1, prob1)))
    #                  else:
    #                      # 'e' -> 's'
    #                      idx0 = idx
    #                      prob0 = prob
    #                  tag0 = tag
    #                  idx0 = idx
    #                  prob0 = prob
    #          if tag0 == 's' and idx0 < len(text_in):
    #              dangling_entity = (idx0 if idx0 == 0 else idx0 - 1, prob0,
    #                                 text_in[idx0 if idx0 == 0 else idx0 - 1:-1])
    #      pairs = [((idx0 if idx0 == 0 else idx0 - 1, prob0), (idx1, prob1))
    #               for (idx0, prob0), (idx1, prob1) in pairs
    #               if idx1 - idx0 >= min_label_len]
    #
    #      entities = [(text_in[idx0:idx1], idx0, idx1, prob0, prob1)
    #                  for (idx0, prob0), (idx1, prob1) in pairs]
    #
    #      return entities, dangling_entity



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .trainer import NerTrainer, load_model, load_pretrained_model, load_pretrained_tokenizer
from .dataset import InputExample, encode_examples, examples_to_dataset
from .dataset import to_BIOS, load_examples_from_bios_file, export_bios_file
from .utils import get_entities
from .args import get_args



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import torch
from torch.autograd import Variable
import torch.nn as nn
import torch.nn.functional as F


def log_sum_exp(vec, m_size):
    """
    Args:
        vec: size=(batch_size, vanishing_dim, hidden_dim)
        m_size: hidden_dim
    Returns:
        size=(batch_size, hidden_dim)
    """
    _, idx = torch.max(vec, 1)  # B * 1 * M
    max_score = torch.gather(vec, 1, idx.view(-1, 1,
                                              m_size)).view(-1, 1,
                                                            m_size)  # B * M
    return max_score.view(-1, m_size) + torch.log(
        torch.sum(torch.exp(vec - max_score.expand_as(vec)), 1)).view(
            -1, m_size)


class CRF(nn.Module):
    #  def __init__(self, **kwargs):
    #      """
    #      Args:
    #          target_size: int, target size
    #          use_cuda: bool, 是否使用gpu, default is True
    #          average_batch: bool, loss是否作平均, default is True
    #      """
    #      super(CRF, self).__init__()
    #      for k in kwargs:
    #          self.__setattr__(k, kwargs[k])
    #      if not hasattr(self, 'average_batch'):
    #          self.__setattr__('average_batch', True)
    #      if not hasattr(self, 'use_cuda'):
    #          self.__setattr__('use_cuda', True)
    #
    #      # init transitions
    #      self.START_TAG_IDX, self.END_TAG_IDX = -2, -1

    def __init__(self, tagset_size, start_tag_idx, stop_tag_idx):
        self.target_size = tagset_size
        self.use_cuda = True
        self.average_batch = True
        self.START_TAG_IDX = start_tag_idx
        self.END_TAG_IDX = stop_tag_idx
        #  init_transitions = torch.zeros(self.target_size + 2,
        #                                 self.target_size + 2)
        init_transitions = torch.zeros(self.target_size, self.target_size)
        init_transitions[:, self.START_TAG_IDX] = -1000.
        init_transitions[self.END_TAG_IDX, :] = -1000.
        if self.use_cuda:
            init_transitions = init_transitions.cuda()
        self.transitions = nn.Parameter(init_transitions)

    def _forward_alg(self, feats, mask):
        """
        Do the forward algorithm to compute the partition function (batched).
        Args:
            feats: size=(batch_size, seq_len, self.target_size+2)
            mask: size=(batch_size, seq_len)
        Returns:
            xxx
        """
        batch_size = feats.size(0)
        seq_len = feats.size(1)
        tag_size = feats.size(-1)

        mask = mask.transpose(1, 0).contiguous()
        ins_num = batch_size * seq_len

        feats = feats.transpose(1, 0).contiguous().view(
            ins_num, 1, tag_size).expand(ins_num, tag_size, tag_size)

        scores = feats + self.transitions.view(1, tag_size, tag_size).expand(
            ins_num, tag_size, tag_size)
        scores = scores.view(seq_len, batch_size, tag_size, tag_size)

        seq_iter = enumerate(scores)
        try:
            _, inivalues = seq_iter.__next__()
        except:
            _, inivalues = seq_iter.next()
        partition = inivalues[:, self.START_TAG_IDX, :].clone().view(
            batch_size, tag_size, 1)

        for idx, cur_values in seq_iter:
            cur_values = cur_values + partition.contiguous().view(
                batch_size, tag_size, 1).expand(batch_size, tag_size, tag_size)
            cur_partition = log_sum_exp(cur_values, tag_size)

            mask_idx = mask[idx, :].view(batch_size,
                                         1).expand(batch_size, tag_size)

            masked_cur_partition = cur_partition.masked_select(mask_idx)
            if masked_cur_partition.dim() != 0:
                mask_idx = mask_idx.contiguous().view(batch_size, tag_size, 1)
                partition.masked_scatter_(mask_idx, masked_cur_partition)

        cur_values = self.transitions.view(1, tag_size, tag_size).expand(
            batch_size, tag_size, tag_size) + partition.contiguous().view(
                batch_size, tag_size, 1).expand(batch_size, tag_size, tag_size)
        cur_partition = log_sum_exp(cur_values, tag_size)
        final_partition = cur_partition[:, self.END_TAG_IDX]
        return final_partition.sum(), scores

    def _viterbi_decode(self, feats, mask):
        """
        Args:
            feats: size=(batch_size, seq_len, self.target_size+2)
            mask: size=(batch_size, seq_len)
        Returns:
            decode_idx: (batch_size, seq_len), viterbi decode结果
            path_score: size=(batch_size, 1), 每个句子的得分
        """
        batch_size = feats.size(0)
        seq_len = feats.size(1)
        tag_size = feats.size(-1)

        length_mask = torch.sum(mask, dim=1).view(batch_size, 1).long()

        mask = mask.transpose(1, 0).contiguous()
        ins_num = seq_len * batch_size

        feats = feats.transpose(1, 0).contiguous().view(
            ins_num, 1, tag_size).expand(ins_num, tag_size, tag_size)

        scores = feats + self.transitions.view(1, tag_size, tag_size).expand(
            ins_num, tag_size, tag_size)
        scores = scores.view(seq_len, batch_size, tag_size, tag_size)

        seq_iter = enumerate(scores)
        # record the position of the best score
        back_points = list()
        partition_history = list()

        # mask = 1 + (-1) * mask
        mask = (1 - mask.long()).byte()
        try:
            _, inivalues = seq_iter.__next__()
        except:
            _, inivalues = seq_iter.next()

        partition = inivalues[:, self.START_TAG_IDX, :].clone().view(
            batch_size, tag_size, 1)
        partition_history.append(partition)

        for idx, cur_values in seq_iter:
            cur_values = cur_values + partition.contiguous().view(
                batch_size, tag_size, 1).expand(batch_size, tag_size, tag_size)
            partition, cur_bp = torch.max(cur_values, 1)
            partition_history.append(partition.unsqueeze(-1))

            cur_bp.masked_fill_(
                mask[idx].view(batch_size, 1).expand(batch_size, tag_size), 0)
            back_points.append(cur_bp)

        partition_history = torch.cat(partition_history).view(
            seq_len, batch_size, -1).transpose(1, 0).contiguous()

        last_position = length_mask.view(batch_size, 1, 1).expand(
            batch_size, 1, tag_size) - 1
        last_partition = torch.gather(partition_history, 1,
                                      last_position).view(
                                          batch_size, tag_size, 1)

        last_values = last_partition.expand(batch_size, tag_size, tag_size) + \
            self.transitions.view(1, tag_size, tag_size).expand(batch_size, tag_size, tag_size)
        _, last_bp = torch.max(last_values, 1)
        pad_zero = Variable(torch.zeros(batch_size, tag_size)).long()
        if self.use_cuda:
            pad_zero = pad_zero.cuda()
        back_points.append(pad_zero)
        back_points = torch.cat(back_points).view(seq_len, batch_size,
                                                  tag_size)

        pointer = last_bp[:, self.END_TAG_IDX]
        insert_last = pointer.contiguous().view(batch_size, 1, 1).expand(
            batch_size, 1, tag_size)
        back_points = back_points.transpose(1, 0).contiguous()

        back_points.scatter_(1, last_position, insert_last)

        back_points = back_points.transpose(1, 0).contiguous()

        decode_idx = Variable(torch.LongTensor(seq_len, batch_size))
        if self.use_cuda:
            decode_idx = decode_idx.cuda()
        decode_idx[-1] = pointer.data
        for idx in range(len(back_points) - 2, -1, -1):
            pointer = torch.gather(back_points[idx], 1,
                                   pointer.contiguous().view(batch_size, 1))
            decode_idx[idx] = pointer.view(-1).data
        path_score = None
        decode_idx = decode_idx.transpose(1, 0)
        return path_score, decode_idx

    def forward(self, feats, mask):
        path_score, best_path = self._viterbi_decode(feats, mask)
        return path_score, best_path

    def _score_sentence(self, scores, mask, tags):
        """
        Args:
            scores: size=(seq_len, batch_size, tag_size, tag_size)
            mask: size=(batch_size, seq_len)
            tags: size=(batch_size, seq_len)
        Returns:
            score:
        """
        batch_size = scores.size(1)
        seq_len = scores.size(0)
        tag_size = scores.size(-1)

        new_tags = Variable(torch.LongTensor(batch_size, seq_len))
        if self.use_cuda:
            new_tags = new_tags.cuda()
        for idx in range(seq_len):
            if idx == 0:
                new_tags[:, 0] = (tag_size - 2) * tag_size + tags[:, 0]
            else:
                new_tags[:, idx] = tags[:, idx - 1] * tag_size + tags[:, idx]

        end_transition = self.transitions[:,
                                          self.END_TAG_IDX].contiguous().view(
                                              1, tag_size).expand(
                                                  batch_size, tag_size)
        length_mask = torch.sum(mask, dim=1).view(batch_size, 1).long()
        end_ids = torch.gather(tags, 1, length_mask - 1)

        end_energy = torch.gather(end_transition, 1, end_ids)

        new_tags = new_tags.transpose(1, 0).contiguous().view(
            seq_len, batch_size, 1)
        tg_energy = torch.gather(scores.view(seq_len, batch_size, -1), 2,
                                 new_tags).view(seq_len, batch_size)
        tg_energy = tg_energy.masked_select(mask.transpose(1, 0))

        gold_score = tg_energy.sum() + end_energy.sum()

        return gold_score

    def neg_log_likelihood_loss(self, feats, mask, tags):
        """
        Args:
            feats: size=(batch_size, seq_len, tag_size)
            mask: size=(batch_size, seq_len)
            tags: size=(batch_size, seq_len)
        """
        batch_size = feats.size(0)
        forward_score, scores = self._forward_alg(feats, mask)
        gold_score = self._score_sentence(scores, mask, tags)
        if self.average_batch:
            return (forward_score - gold_score) / batch_size
        return forward_score - gold_score



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# ----------------------------------------
"""
扩增训练样本数据
随机替换训练数据标签，增加训练样本。
"""

import re
import numpy as np


def findall_substring(text: str, s: str) -> list:
    """
    查找text中所有匹配的文本片断s（可以是正则表达式）
    返回列表 [(start_pos, end_pos, substring, [regex_group1, regex_group2, ...])]

    """

    if s[:1] != "(":
        s = f"({s})"

    mobjs = re.finditer(s, text)
    results = [(m.start(), m_end(), m.group(), m.groups()) for m in mobjs]
    return results


def get_label_frame(text: str, label_positions: list) -> list:
    """
    获取文本的标注填充框架
    label_positions: 填充标记的起止位置对列表。
    返回扣除标记后的文本片断列表。
    例子：
        frame = get_label_frame("我去北京上学", [(2,4)])
        assert frame == ["我去", "上学"]
    """
    assert text and label_positions

    frame = []
    i0 = 0
    for s, e in label_positions:
        frame.append(text[i0:s])
        i0 = e
    frame.append(text[i0:])

    assert len(frame) >= 2
    assert len(frame) == len(label_positions) + 1

    return frame


def get_frame_entities_count(frame: list) -> int:
    """
    返回实体填充模板要求的实体个数
    例子：
        frame = get_label_frame("我去北京上学", [(2,4)])
        num_entities = get_frame_entities_count(frame)
        assert num_entities == 1
    """
    if len(frame) < 2:
        return 0
    else:
        return len(frame) - 1


def fill_frame(frame: list, labels: list):
    """
    填充实体标注模板
    模板要求填充实体个数可调用get_frame_entities_count(frame)获得，
    参数labels列表条目个数必须等于模板要求个数。
    例子：
        frame = get_label_frame("我去北京上学", [(2,4)])
        text = fill_frame(frame, ['上海'])
        assert text == "我去上海上学"
    """
    assert len(list) == len(labels) + 1

    text = "".join(frame_text + label
                   for frame_text, label in zip(frame, labels + [""]))
    return text


# 从训练数据集中获得所有标注文本
def get_all_labels(train_data):
    labels = list(set([label for ID, text, category, label in train_data]))
    return labels


# -------- 获取所有扣除标签的样本框架 --------
def get_all_frames(train_data):
    sample_frames = []
    for ID, text, category, label in tqdm(train_data, desc="Frames"):
        assert len(text) > 0 and len(label) > 0

        found_labels = findall_substring(text, label)
        if len(found_labels) > 0:
            s, e, _, _ = found_labels[0]
            frame = get_label_frame(text, [(s, e)])
            sample_frames.append(frame)
    assert len(sample_frames) == len(train_data)

    return sample_frames


# 实现随机替换标注实体，按标注模板生成新的训练样本。
#  labels = get_all_labels(train_data)
#  frames = get_all_frames(train_data)
def generate_fake_train_data(train_data, frames, labels):

    assert len(labels) == len(frames)

    fake_train_data = []
    for frame, (ID, _, category, _) in tqdm(zip(sample_frames, train_data),
                                     desc="Generate fake"):
        np.random.shuffle(labels)
        picked_label = labels[0]

        fake_text = fill_frame(frame, [picked_label])

        fake_train_data.append(
            (f"{ID}_{n}", fake_text, category, picked_labels))

    return fake_train_data


# ----------------------------------------
def augment_train_data(train_data: list, num_augments=1) -> list:
    aug_train_data = []

    if len(train_data) > 0 and num_augments > 0:
        # -------- 获取所有训练标签 --------
        labels = list(set([label for ID, text, category, label in train_data]))
        assert len(labels) > 0

        categories = list(set([c for _, _, c, _ in train_data]))
        assert len(categories) == 1
        category = categories[0]

        # -------- 获取所有扣除标签的样本框架 --------
        sample_frames = []
        for ID, text, category, label in tqdm(train_data, desc="Frames"):
            assert len(text) > 0 and len(label) > 0

            found_labels = findall_substring(text, label)
            if len(found_labels) > 0:
                s, e, _, _ = found_labels[0]
                frame = get_label_frame(text, [(s, e)])
                sample_frames.append(frame)
        assert len(sample_frames) == len(train_data)

        for frame, (ID, _, _, _) in tqdm(zip(sample_frames, train_data),
                                         desc="Augementing"):
            for n in range(num_augments):
                np.random.shuffle(labels)
                picked_labels = [labels[0]]

                assert get_frame_entities_count(frame) == len(picked_labels)

                fake_text = fill_frame(frame, picked_labels)

                aug_train_data.append(
                    (f"{ID}_{n}", fake_text, category, picked_labels[0]))

    return aug_train_data



import torch
import torch.nn as nn
import torch.nn.functional as F


def to_scalar(var):
    return var.view(-1).detach().tolist()[0]


def argmax(vec):
    _, idx = torch.max(vec, 1)
    return to_scalar(idx)


def log_sum_exp(vec):
    max_score = vec[0, argmax(vec)]
    max_score_broadcast = max_score.view(1, -1).expand(1, vec.size()[1])
    return max_score + torch.log(
        torch.sum(torch.exp(vec - max_score_broadcast)))


def argmax_batch(vecs):
    _, idx = torch.max(vecs, 1)
    return idx


def log_sum_exp_batch(vecs):
    maxi = torch.max(vecs, 1)[0]
    maxi_bc = maxi[:, None].repeat(1, vecs.shape[1])
    recti_ = torch.log(torch.sum(torch.exp(vecs - maxi_bc), 1))
    return maxi + recti_


class CRF(nn.Module):
    #  def __init__(self, tagset_size, tag_dictionary, device, is_bert=True):
    #      super(CRF, self).__init__()
    #
    #      self.START_TAG = "<START>"
    #      self.STOP_TAG = "<STOP>"
    #      if is_bert:
    #          self.START_TAG = "[CLS]"
    #          self.STOP_TAG = "[SEP]"
    #      self.tag_dictionary = tag_dictionary
    def __init__(self, tagset_size, start_tag_idx, stop_tag_idx):
        super(CRF, self).__init__()
        self.tagset_size = tagset_size
        self.START_TAG_IDX = start_tag_idx
        self.STOP_TAG_IDX = stop_tag_idx

        self.device = torch.device(
            "cuda" if torch.cuda.is_available() else "cpu")
        self.transitions = torch.randn(tagset_size, tagset_size)
        # self.transitions = torch.zeros(tagset_size, tagset_size)
        self.transitions.detach()[self.START_TAG_IDX, :] = -10000
        #  self.tag_dictionary[self.START_TAG], :] = -10000
        self.transitions.detach()[:, self.STOP_TAG_IDX] = -10000
        #  self.tag_dictionary[self.STOP_TAG]] = -10000
        self.transitions = self.transitions.to(self.device)
        self.transitions = nn.Parameter(self.transitions)

    def _viterbi_decode(self, feats):
        backpointers = []
        backscores = []
        scores = []
        init_vvars = (torch.FloatTensor(1, self.tagset_size).to(
            self.device).fill_(-10000.0))
        #  init_vvars[0][self.tag_dictionary[self.START_TAG]] = 0
        init_vvars[0][self.START_TAG_IDX] = 0
        forward_var = init_vvars

        for feat in feats:
            next_tag_var = (forward_var.view(1, -1).expand(
                self.tagset_size, self.tagset_size) + self.transitions)
            _, bptrs_t = torch.max(next_tag_var, dim=1)
            viterbivars_t = next_tag_var[range(len(bptrs_t)), bptrs_t]
            forward_var = viterbivars_t + feat
            backscores.append(forward_var)
            backpointers.append(bptrs_t)

        #  terminal_var = (forward_var +
        #                  self.transitions[self.tag_dictionary[self.STOP_TAG]])
        #  terminal_var.detach()[self.tag_dictionary[self.STOP_TAG]] = -10000.0
        #  terminal_var.detach()[self.tag_dictionary[self.START_TAG]] = -10000.0
        terminal_var = (forward_var +
                        self.transitions[self.STOP_TAG_IDX])
        terminal_var.detach()[self.STOP_TAG_IDX] = -10000.0
        terminal_var.detach()[self.START_TAG_IDX] = -10000.0
        best_tag_id = argmax(terminal_var.unsqueeze(0))
        best_path = [best_tag_id]
        for bptrs_t in reversed(backpointers):
            best_tag_id = bptrs_t[best_tag_id]
            best_path.append(best_tag_id.item())
        best_scores = []
        for backscore in backscores:
            softmax = F.softmax(backscore, dim=0)
            _, idx = torch.max(backscore, 0)
            prediction = idx.item()
            best_scores.append(softmax[prediction].item())
            scores.append([elem.item() for elem in softmax.flatten()])
        swap_best_path, swap_max_score = (
            best_path[0],
            scores[-1].index(max(scores[-1])),
        )
        scores[-1][swap_best_path], scores[-1][swap_max_score] = (
            scores[-1][swap_max_score],
            scores[-1][swap_best_path],
        )
        start = best_path.pop()
        #  assert start == self.tag_dictionary[self.START_TAG]
        assert start == self.START_TAG_IDX
        best_path.reverse()
        return best_scores, best_path, scores

    def _forward_alg(self, feats, lens_):
        init_alphas = torch.FloatTensor(self.tagset_size).fill_(-10000.0)
        #  init_alphas[self.tag_dictionary[self.START_TAG]] = 0.0
        init_alphas[self.START_TAG_IDX] = 0.0

        forward_var = torch.zeros(
            feats.shape[0],
            feats.shape[1] + 1,
            feats.shape[2],
            dtype=torch.float,
            device=self.device,
        )
        forward_var[:, 0, :] = init_alphas[None, :].repeat(feats.shape[0], 1)
        transitions = self.transitions.view(1, self.transitions.shape[0],
                                            self.transitions.shape[1]).repeat(
                                                feats.shape[0], 1, 1)
        for i in range(feats.shape[1]):
            emit_score = feats[:, i, :]
            tag_var = (
                emit_score[:, :, None].repeat(1, 1, transitions.shape[2]) +
                transitions + forward_var[:, i, :][:, :, None].repeat(
                    1, 1, transitions.shape[2]).transpose(2, 1))
            max_tag_var, _ = torch.max(tag_var, dim=2)
            tag_var = tag_var - max_tag_var[:, :, None].repeat(
                1, 1, transitions.shape[2])
            agg_ = torch.log(torch.sum(torch.exp(tag_var), dim=2))
            cloned = forward_var.clone()
            cloned[:, i + 1, :] = max_tag_var + agg_
            forward_var = cloned
        forward_var = forward_var[range(forward_var.shape[0]), lens_, :]
        terminal_var = forward_var + self.transitions[self.tag_dictionary[
            self.STOP_TAG]][None, :].repeat(forward_var.shape[0], 1)
        alpha = log_sum_exp_batch(terminal_var)
        return alpha

    def _score_sentence(self, feats, tags, lens_):
        #  start = torch.LongTensor([self.tag_dictionary[self.START_TAG]
        #                            ]).to(self.device)
        start = torch.LongTensor([self.START_TAG_IDX]).to(self.device)
        start = start[None, :].repeat(tags.shape[0], 1)
        #  stop = torch.LongTensor([self.tag_dictionary[self.STOP_TAG]
        #                           ]).to(self.device)
        stop = torch.LongTensor([self.STOP_TAG_IDX]).to(self.device)
        stop = stop[None, :].repeat(tags.shape[0], 1)
        pad_start_tags = torch.cat([start, tags], 1)
        pad_stop_tags = torch.cat([tags, stop], 1)
        for i in range(len(lens_)):
            #  pad_stop_tags[i, lens_[i]:] = self.tag_dictionary[self.STOP_TAG]
            pad_stop_tags[i, lens_[i]:] = self.STOP_TAG_IDX
        score = torch.FloatTensor(feats.shape[0]).to(self.device)
        for i in range(feats.shape[0]):
            r = torch.LongTensor(range(lens_[i])).to(self.device)
            score[i] = torch.sum(self.transitions[
                pad_stop_tags[i, :lens_[i] + 1],
                pad_start_tags[i, :lens_[i] + 1]]) + torch.sum(
                    feats[i, r, tags[i, :lens_[i]]])
        return score

    def obtain_labels(self, feature, input_lens):
        tags = []
        all_tags = []
        for feats, length in zip(feature, input_lens):
            confidences, tag_seq, scores = self._viterbi_decode(feats[:length])
            tags.append([tag for tag in tag_seq])
            all_tags.append(
                [[score_id for score_id, score in enumerate(score_dist)]
                 for score_dist in scores])
        return tags, all_tags

    def _obtain_labels(self, feature, id2label, input_lens):
        tags = []
        all_tags = []
        for feats, length in zip(feature, input_lens):
            confidences, tag_seq, scores = self._viterbi_decode(feats[:length])
            tags.append([id2label[tag] for tag in tag_seq])
            all_tags.append([[
                id2label[score_id] for score_id, score in enumerate(score_dist)
            ] for score_dist in scores])
        return tags, all_tags

    def calculate_loss(self, scores, tag_list, lengths):
        return self._calculate_loss_old(scores, lengths, tag_list)

    def _calculate_loss_old(self, features, lengths, tags):
        forward_score = self._forward_alg(features, lengths)
        gold_score = self._score_sentence(features, tags, lengths)
        score = forward_score - gold_score
        return score.mean()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import random, copy, json
from tqdm import tqdm
from loguru import logger
from theta.utils import seg_generator
from dataclasses import dataclass, field

from ..utils import seg_generator


#  @dataclass(frozen=False)
class InputExample:
    """
    A single training/test example for simple sequence classification.

    Args:
        guid: Unique id for the example.
        text_a: string. The untokenized text of the first sequence. For single
            sequence tasks, only this sequence must be specified.
        text_b: (Optional) string. The untokenized text of the second sequence.
            Only must be specified for sequence pair tasks.
        label: (Optional) string. The label of the example. This should be
            specified for train and dev examples, but not for test examples.
    """

    #  guid: str
    #  text_a: str
    #  text_b: Optional[str] = None
    #  label: Optional[str] = None
    def __init__(self, guid, text_a, text_b=None, label=None):
        self.guid = guid
        self.text_a = text_a
        self.text_b = text_b
        self.label = label

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        return json.dumps(dataclasses.asdict(self), indent=2,
                          sort_keys=True) + "\n"


def load_glue_examples(data_generator, examples_file):

    examples = []

    for guid, text_a, text_b, label in data_generator(examples_file):
        examples.append(InputExample(guid=guid, text_a=text_a, text_b = text_b, label=label))
    logger.info(f"Loaded {len(examples)} examples.")

    return examples


def show_glue_datainfo(glue_labels, train_data_generator, train_file,
                       test_data_generator, test_file):
    train_lengths = [
        len(text_a)
        for guid, text_a, _, label in train_data_generator(train_file)
    ]

    all_labels = []
    for _, _, _, label in train_data_generator(train_file):
        all_labels.append(label)
        if label not in glue_labels:
            glue_labels.append(label)

    test_lengths = [
        len(text_a)
        for guid, text_a, _, label in test_data_generator(test_file)
    ]

    logger.info(f"****** glue_labels ******")
    logger.info(f"{glue_labels}")
    from collections import Counter
    logger.info(f"{Counter(all_labels).most_common()}")
    logger.info(f"")
    import numpy as np
    logger.info(f"****** train lengths ******")
    logger.info(f"mean: {np.mean(train_lengths):.2f}")
    logger.info(f"std: {np.std(train_lengths):.2f}")
    logger.info(f"min: {np.min(train_lengths)}")
    logger.info(f"max: {np.max(train_lengths)}")
    logger.info(f"")

    logger.info(f"****** test lengths ******")
    logger.info(f"mean: {np.mean(test_lengths):.2f}")
    logger.info(f"std: {np.std(test_lengths):.2f}")
    logger.info(f"min: {np.min(test_lengths)}")
    logger.info(f"max: {np.max(test_lengths)}")
    logger.info(f"")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import torch
from torch import nn
import torch.nn.functional as F


def swish(x):
    return x * torch.sigmoid(x)


class Swish(nn.Module):

    def __init__(self):
        '''
        Init method.
        '''
        super().__init__()

    def forward(self, x):
        '''
        Forward pass of the function.
        '''
        return swish(x)



# -*- coding: utf-8 -*-


#%% ------------------------------------------------------------
# TPU
def tpu_available():
    import os
    return 'COLAB_TPU_ADDR' in os.environ


def init_tpu():
    strategy = None
    if tpu_available():
        print(f"TPU available.")
        tf.logging.set_verbosity(tf.logging.INFO)
        resolver = tf.distribute.cluster_resolver.TPUClusterResolver()
        tf.config.experimental_connect_to_host(resolver.master())
        tf.tpu.experimental.initialize_tpu_system(resolver)
        strategy = tf.distribute.experimental.TPUStrategy(resolver)
    return strategy


def convert_data_to_tensors(bert_data):
    X_data = []
    num_steps = len(bert_data)
    print(f"{num_steps} steps.")
    i = 0
    with tqdm(total=num_steps) as pbar:
        for (X1, X2, S1, S2), _ in bert_data.__iter__(max_length):
            if i >= num_steps:
                break
            for x1, x2, s1, s2 in zip(X1, X2, S1, S2):
                X_data.append([x1, x2, s1, s2])
            pbar.update(1)
            i += 1
    return X_data


def train_model_on_tpu(train_data, dev_data):
    print(f"TPU available.")
    batch_size = 128
    tpu_cores = 8
    bert_train_data = BertTrainData(train_data,
                                    max_length,
                                    batch_size=batch_size)
    bert_dev_data = BertTrainData(dev_data,
                                  extract_max_length,
                                  batch_size=batch_size)

    X_train = convert_data_to_tensors(bert_train_data)
    print(f"\nX_train len: {len(X_train)}")
    X_dev = convert_data_to_tensors(bert_dev_data)
    print(f"\nX_dev {len(X_dev)}")

    X_train = []
    num_steps = len(bert_train_data)
    print(f"{num_steps} steps.")
    i = 0
    with tqdm(total=num_steps) as pbar:
        for (X1, X2, S1, S2), _ in bert_train_data.__iter__(max_length):
            if i >= num_steps:
                break
            #print(f"X1.shape: {X1.shape} X2.shape: {X2.shape} S1.shape: {S1.shape} S2.shape: {S2.shape}")
            #X = np.concatenate([X1, X2, S1, S2], axis=1).reshape((-1, 4, X1.shape[1]))
            #X_train.append(list(X))
            for x1, x2, s1, s2 in zip(X1, X2, S1, S2):
                X_train.append([x1, x2, s1, s2])
            pbar.update(1)
            i += 1

    print(f"\nX_train.shape: {len(X_train)}")

    print(f"batch_size: {batch_size} tpu_cores: {tpu_cores}")
    strategy = init_tpu()
    with strategy.scope():
        X_train = convert_data_to_tensors(bert_train_data)
        print(f"\nX_train len: {len(X_train)}")
        X_dev = convert_data_to_tensors(bert_dev_data)
        print(f"\nX_dev {len(X_dev)}")
        train_model.fit(
            X_train,
            steps_per_epoch=len(bert_train_data) / tpu_cores,
            #steps_per_epoch=int(len(bert_train_data) / (batch_size * tpu_cores)),
            epochs=8,
            batch_size=batch_size * 8,
            validation_data=(X_dev, 0))


#  X_train = convert_data_to_tensors(bert_train_data)
#  print(f"\nX_train len: {len(X_train)}")
#  X_dev = convert_data_to_tensors(bert_dev_data)
#  print(f"\nX_dev {len(X_dev)}")
#  bert_model.train_model.fit(
#      X_train,
#      #  steps_per_epoch=len(bert_train_data) / tpu_cores,
#      #steps_per_epoch=int(len(bert_train_data) / (batch_size * tpu_cores)),
#      epochs=3,
#      batch_size=batch_size * gpus,
#      callbacks=[evaluator])

#  if tpu_available():
#      self.on_epoch_begin_tpu(batch, logs)
#      return

#  def get_lr_opt(self):
#      if not hasattr(self, 'model'):
#          return None, None
#      tmp = self.model
#      lr_attr_name = None
#      lr_name_list = ['lr', '_lr', '_learning_rate', 'learning_rate']
#      opt_name_list = ['optimizer', '_opt']
#      while True:
#          for lr_name in lr_name_list:
#              if hasattr(tmp, lr_name):
#                  lr_attr_name = lr_name
#                  break
#          if lr_attr_name != None:
#              return tmp, lr_attr_name
#          for opt_name in opt_name_list:
#              if hasattr(tmp, opt_name):
#                  tmp = getattr(tmp, opt_name)
#                  break
#      return None, None
#
#  def on_epoch_begin_tpu(self, epoch, logs=None):
#      opt, lr_name = self.get_lr_opt()
#      if opt is None:
#          raise ValueError('no attr _lr, _learning or lr found')
#      try:  # new API
#          lr = getattr(opt, lr_name)
#          lr = self.schedule(epoch, lr)
#      except TypeError:  # Support for old API for backward compatibility
#          lr = self.schedule(epoch)
#      if not isinstance(lr, (float, np.float32, np.float64)):
#          raise ValueError('The output of the "schedule" function '
#                           'should be float.')
#      setattr(opt, lr_name, lr)
#      if self.verbose > 0:
#          print('\nEpoch %05d: LearningRateScheduler reducing learning '
#                'rate to %s.' % (epoch + 1, lr))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .trainer import SenembTrainer, load_pretrained_tokenizer, load_pretrained_model
from .dataset import InputExample, examples_to_dataset, build_encode_dataset



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os, random, json
import numpy as np
from tqdm import tqdm
from loguru import logger
import math
import torch
from sklearn.metrics import f1_score
from .multiprocesses import is_single_process


def load_pytorch_model(model, model_path):
    if hasattr(model, "module"):
        model.module.load_state_dict(torch.load(model_path,
                                                map_location='cpu'),
                                     strict=False)
    else:
        model.load_state_dict(torch.load(model_path, map_location='cpu'),
                              strict=False)
    return model


def softmax(x):
    e_x = np.exp(x - np.max(x, axis=1).reshape(-1, 1))
    return e_x / e_x.sum(axis=1).reshape(-1, 1)


def sigmoid(x):
    return 1 / (1 + np.exp(-x))


def simple_accuracy(preds, labels):
    return (preds == labels).mean()


def acc_and_f1(preds, labels):
    correct = np.sum((preds == labels).astype(int))
    acc = correct / preds.shape[0]
    f1 = f1_score(y_true=labels, y_pred=preds, average="macro")
    acc_and_f1 = (acc + f1) / 2
    return {"acc": acc, "f1": f1, "acc_and_f1": acc_and_f1}


def init_random(seed):

    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)

    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True


# Setup CUDA, GPU & distributed training
def init_cuda(args):
    #  if args.local_rank == -1 or args.no_cuda:
    if is_single_process(args) or args.no_cuda:
        device = torch.device("cuda" if torch.cuda.is_available()
                              and not args.no_cuda else "cpu")
        args.n_gpu = torch.cuda.device_count()
    else:
        # Initializes the distributed backend which will
        # take care of sychronizing nodes/GPUs
        torch.cuda.set_device(args.local_rank)
        device = torch.device("cuda", args.local_rank)
        torch.distributed.init_process_group(backend="nccl")
        args.n_gpu = 1
    args.train_batch_size = args.per_gpu_train_batch_size * max(1, args.n_gpu)
    args.eval_batch_size = args.per_gpu_eval_batch_size * max(1, args.n_gpu)
    args.device = device


def init_theta(args):
    init_random(args.seed)
    init_cuda(args)


# DataFrame直接初始化出InputExample列表
def dataframe_to_examples(df_data, InputExample, cols_map=None):
    if cols_map is None:
        cols_map = {x: x for x in df_data.columns()}
    examples = []
    for i, row in df_data.iterrows():
        values = {k: str(row[v]) for k, v in cols_map.items()}
        examples.append(InputExample(**values))

    return np.array(examples)


def get_list_size(samples: [list, np.array]):
    if isinstance(samples, list):
        num_samples = len(samples)
    elif isinstance(samples, np.ndarray):
        num_samples = samples.shape[0]
    else:
        raise TypeError(
            f"Argument samples in shuffle_list() must be a list or np.ndarray."
        )
    return num_samples


def shuffle_list(samples: [list, np.array],
                 random_state=None) -> [list, np.array]:
    if random_state:
        np.random.seed(random_state)

    num_samples = get_list_size(samples)
    indices = np.random.randint(0, num_samples, num_samples)
    shuffled_samples = [samples[i] for i in indices]

    if isinstance(samples, np.ndarray):
        return np.array(shuffled_samples)
    else:
        return shuffled_samples


def list_to_list(samples: [list, np.array]) -> list:
    if isinstance(samples, np.ndarray):
        samples = samples.tolist()
        return samples
    elif isinstance(samples, list):
        return samples
    else:
        raise TypeError(
            f"Samples in list_to_list() must be a list or np.ndarray.")


def concatenate_list(a_samples: [list, np.array],
                     b_samples: [list, np.array]) -> [list, np.array]:
    if isinstance(a_samples, np.ndarray):
        a_samples = list_to_list(a_samples)
        b_samples = list_to_list(b_samples)
        samples = a_samples + b_samples
        return np.array(samples)
    elif isinstance(a_samples, list):
        a_samples = list_to_list(a_samples)
        b_samples = list_to_list(b_samples)
        samples = a_samples + b_samples
        return samples
    else:
        raise TypeError(
            f"Samples in concatenote_list() must be a list or np.ndarray.")


def to_numpy(X):
    """
    Convert input to numpy ndarray
    """
    if hasattr(X, 'iloc'):  # pandas
        return X.values
    elif isinstance(X, list):  # list
        return np.array(X)
    elif isinstance(X, np.ndarray):  # ndarray
        return X
    else:
        raise ValueError("Unable to handle input type %s" % str(type(X)))


def unpack_text_pairs(X):
    assert isinstance(X, list) or isinstance(X, np.ndarray)
    return (X, None) if X.ndim == 1 else (X[:, 0], X[:, 1])


def unpack_data(X, y=None):
    return unpack_text_pairs(to_numpy(X)), to_numpy(y) if y else None


def create_logger(log_dir: str, logger_name: str):
    # 创建一个logger
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)

    if not os.path.exists(log_dir):
        os.mkdir(log_dir)

    # 创建一个handler，
    timestamp = time.strftime("%Y.%m.%d_%H.%M.%S", time.localtime())
    log_filename = f"{log_dir}/log_{logger_name}_{timestamp}.txt"
    fh = logging.FileHandler(log_filename)
    fh.setLevel(logging.DEBUG)

    # 再创建一个handler，用于输出到控制台
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)

    # 定义handler的输出格式
    formatter = logging.Formatter(
        '[%(asctime)s][%(levelname)s] ## %(message)s')
    fh.setFormatter(formatter)
    ch.setFormatter(formatter)
    # 给logger添加handler
    logger.addHandler(fh)
    logger.addHandler(ch)

    return logger


def seg_generator(iterables, seg_len, seg_backoff=0):
    if seg_len <= 0:
        yield iterables, 0
    else:
        #  # 确保iterables列表中每一项的条目数相同
        #  assert sum([len(x)
        #              for x in iterables]) == len(iterables[0]) * len(iterables)
        assert iterables[0] is not None
        s0 = 0
        while s0 < len(iterables[0]):
            s1 = s0 + seg_len
            segs = [x[s0:s1] if x else None for x in iterables]
            yield segs, s0
            s0 += seg_len - seg_backoff


class slide_generator:
    def __init__(self, words: list, slide_len, slide_backoff=0):

        assert slide_offset <= slide_len

        self.words = words
        self.slide_len = slide_len
        self.slide_backoff = slide_backoff

        num_words = len(words)
        self.total_slides = num_words // (slide_len - slide_backoff) + 1

    def __iter__(self):
        for i in range(self.total_slides):
            yield self.__getitem__(i)

    def __len__(self):
        return self.total_slides

    def __getitem__(self, idx):
        s = (self.slide_len - self.slide_backoff) * idx
        e = (self.slide_len - self.slide_backoff) * (idx + 1)
        slide_words = self.words[s:e]

        return slide_words


def split_train_eval_examples(examples: [list, np.array],
                              train_rate=0.9,
                              fold=0,
                              shuffle=False,
                              random_state=None) -> [list, np.array]:
    examples = list(examples)
    num_examples = len(examples)

    # for entity_typing
    if shuffle:
        random.shuffle(examples)

    num_train_examples = int(num_examples * train_rate)
    num_eval_examples = num_examples - num_train_examples

    e = num_examples - num_eval_examples * fold
    s = num_examples - num_eval_examples * (fold + 1)
    if s < 0:
        s = 0
    eval_examples = examples[s:e]
    train_examples = concatenate_list(examples[:s], examples[e:])

    # for medical_entity
    #  if shuffle:
    #      examples = shuffle_list(examples, random_state=random_state)
    #
    #  num_eval_examples = int(num_examples * (1 - train_rate))
    #  assert fold <= num_examples // num_eval_examples
    #
    #  s = num_eval_examples * fold
    #  e = num_eval_examples * (fold + 1)
    #  eval_examples = examples[s:e]
    #  train_examples = concatenate_list(examples[:s], examples[e:])

    return train_examples, eval_examples


class batch_generator:
    def __init__(self,
                 examples: [list, np.array],
                 train_rate=0.9,
                 shuffle=False,
                 random_state=None):
        if shuffle:
            examples = shuffle_list(examples, random_state=random_state)
        num_examples = get_list_size(examples)
        num_eval_examples = int(num_examples * (1 - train_rate))

        self.examples = examples
        self.total_batchs = num_examples // num_eval_examples + 1

    def __iter__(self):
        for i in range(self.total_batchs):
            yield self.__getitem__(i)

    def __getitem__(self, idx):
        s = self.num_eval_examples * idx
        e = self.num_eval_examples * (idx + 1)
        eval_examples = self.examples[s:e]
        train_examples = concatenate_list(self.examples[:s], self.examples[e:])
        return train_examples, eval_examples

    def __len__(self):
        return self.total_batchs


def get_pred_results_file(args):
    return f"{args.output_dir}/{args.dataset_name}_result.json"


def get_submission_file(args):
    return f"{args.output_dir}/{args.dataset_name}_submission.json"


def load_json_file(json_file):
    rd = open(json_file, 'r')
    lines = rd.readlines()
    rd.close()
    json_data = []
    for line in tqdm(lines):
        line = line.strip()
        line_data = json.loads(line)
        json_data.append(line_data)
    print(f"Total: {len(json_data)}")
    print(json_data[:5])
    return json_data


def show_statistical_distribution(data_list):
    logger.info(f"***** Statistical Distribution *****")
    logger.info(f"mean: {np.mean(data_list):.4f}")
    logger.info(f"std: {np.mean(data_list):.4f}")
    logger.info(f"max: {np.max(data_list):.4f}")
    logger.info(f"min: {np.min(data_list)}:.4f")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# https://github.com/mali19064/LSTM-CRF-pytorch-faster/blob/master/LSTM_CRF_faster.py

import torch
import torch.nn as nn
import torch.optim as optim
#  START_TAG = "<START>"
#  STOP_TAG = "<STOP>"
#  EMBEDDING_DIM = 5
#  HIDDEN_DIM = 4


def prepare_sequence(seq, to_ix):
    idxs = [to_ix[w] for w in seq]
    return torch.tensor(idxs, dtype=torch.long)


class CRF(nn.Module):

    #  def __init__(self, vocab_size, tag_to_ix, embedding_dim, hidden_dim):
    def __init__(self, tagset_size, start_tag_idx, stop_tag_idx):
        super(CRF, self).__init__()
        #  self.embedding_dim = embedding_dim
        #  self.hidden_dim = hidden_dim
        #  self.vocab_size = vocab_size
        #  self.tag_to_ix = tag_to_ix
        #  self.tagset_size = len(tag_to_ix)
        self.tagset_size = tagset_size + 2
        self.START_TAG_IDX = start_tag_idx
        self.STOP_TAG_IDX = stop_tag_idx
        #  self.tag_to_ix = {START_TAG: -1, STOP_TAG: -2}

        #  self.word_embeds = nn.Embedding(vocab_size, embedding_dim)
        #  self.lstm = nn.LSTM(embedding_dim,
        #                      hidden_dim // 2,
        #                      num_layers=1,
        #                      bidirectional=True)
        #
        #  # Maps the output of the LSTM into tag space.
        #  self.hidden2tag = nn.Linear(hidden_dim, self.tagset_size)

        # Matrix of transition parameters.  Entry i,j is the score of
        # transitioning *to* i *from* j.
        #  self.transitions = nn.Parameter(
        #      torch.randn(self.tagset_size, self.tagset_size))
        #
        #  # These two statements enforce the constraint that we never transfer
        #  # to the start tag and we never transfer from the stop tag
        #  self.transitions.data[self.tag_to_ix[START_TAG], :] = -10000
        #  self.transitions.data[:, self.tag_to_ix[STOP_TAG]] = -10000
        #
        #  self.transitions = self.transitions.cuda()

        init_transitions = torch.randn(self.tagset_size, self.tagset_size)

        # These two statements enforce the constraint that we never transfer
        # to the start tag and we never transfer from the stop tag
        #  init_transitions.data[self.tag_to_ix[START_TAG], :] = -10000
        #  init_transitions.data[:, self.tag_to_ix[STOP_TAG]] = -10000
        init_transitions.data[self.START_TAG_IDX, :] = -10000
        init_transitions.data[:, self.STOP_TAG_IDX] = -10000

        self.transitions = nn.Parameter(init_transitions.cuda())

    #      self.hidden = self.init_hidden()
    #
    #  def init_hidden(self):
    #      return (torch.randn(2, 1, self.hidden_dim // 2),
    #              torch.randn(2, 1, self.hidden_dim // 2))

    def _forward_alg(self, feats):
        # Do the forward algorithm to compute the partition function
        init_alphas = torch.full([self.tagset_size], -10000.)
        # START_TAG has all of the score.
        #  init_alphas[self.tag_to_ix[START_TAG]] = 0.
        init_alphas[self.START_TAG_IDX] = 0.

        # Wrap in a variable so that we will get automatic backprop
        # Iterate through the sentence
        forward_var_list = []
        forward_var_list.append(init_alphas)
        for feat_index in range(feats.shape[0]):
            gamar_r_l = torch.stack([forward_var_list[feat_index]] *
                                    feats.shape[1])
            t_r1_k = torch.unsqueeze(feats[feat_index], 0).transpose(0, 1)
            #  aa = gamar_r_l.cpu() + t_r1_k.cpu() + self.transitions
            aa = gamar_r_l.cuda() + t_r1_k + self.transitions
            forward_var_list.append(torch.logsumexp(aa, dim=1))
        #  terminal_var = forward_var_list[-1] + self.transitions[
        #      self.tag_to_ix[STOP_TAG]]
        terminal_var = forward_var_list[-1] + self.transitions[
            self.STOP_TAG_IDX]
        terminal_var = torch.unsqueeze(terminal_var, 0)
        alpha = torch.logsumexp(terminal_var, dim=1)[0]
        return alpha

    #  def _get_lstm_features(self, sentence):
    #      self.hidden = self.init_hidden()
    #      embeds = self.word_embeds(sentence).view(len(sentence), 1, -1)
    #      lstm_out, self.hidden = self.lstm(embeds, self.hidden)
    #      lstm_out = lstm_out.view(len(sentence), self.hidden_dim)
    #      lstm_feats = self.hidden2tag(lstm_out)
    #      return lstm_feats
    #
    def _score_sentence(self, feats, tags):
        # Gives the score of a provided tag sequence
        score = torch.zeros(1)
        tags = torch.cat([
            #  torch.tensor([self.tag_to_ix[START_TAG]], dtype=torch.long), tags
            torch.tensor([self.START_TAG_IDX], dtype=torch.long),
            tags
        ])
        for i, feat in enumerate(feats):
            score = score + \
                self.transitions[tags[i + 1], tags[i]] + feat[tags[i + 1]]
        #  score = score + self.transitions[self.tag_to_ix[STOP_TAG], tags[-1]]
        score = score + self.transitions[self.STOP_TAG_IDX, tags[-1]]
        return score

    def _viterbi_decode(self, feats):
        backpointers = []
        # Initialize the viterbi variables in log space
        init_vvars = torch.full((1, self.tagset_size), -10000.)
        #  init_vvars[0][self.tag_to_ix[START_TAG]] = 0
        init_vvars[0][self.START_TAG_IDX] = 0

        # forward_var at step i holds the viterbi variables for step i-1
        forward_var_list = []
        forward_var_list.append(init_vvars)

        for feat_index in range(feats.shape[0]):
            gamar_r_l = torch.stack([forward_var_list[feat_index]] *
                                    feats.shape[1])
            gamar_r_l = torch.squeeze(gamar_r_l)
            next_tag_var = gamar_r_l + self.transitions
            viterbivars_t, bptrs_t = torch.max(next_tag_var, dim=1)

            t_r1_k = torch.unsqueeze(feats[feat_index], 0)
            forward_var_new = torch.unsqueeze(viterbivars_t, 0) + t_r1_k

            forward_var_list.append(forward_var_new)
            backpointers.append(bptrs_t.tolist())

        # Transition to STOP_TAG
        #  terminal_var = forward_var_list[-1] + self.transitions[
        #      self.tag_to_ix[STOP_TAG]]
        terminal_var = forward_var_list[-1] + self.transitions[
            self.STOP_TAG_IDX]
        best_tag_id = torch.argmax(terminal_var).tolist()
        path_score = terminal_var[0][best_tag_id]

        # Follow the back pointers to decode the best path.
        best_path = [best_tag_id]
        for bptrs_t in reversed(backpointers):
            best_tag_id = bptrs_t[best_tag_id]
            best_path.append(best_tag_id)
        # Pop off the start tag (we dont want to return that to the caller)
        start = best_path.pop()
        #  assert start == self.tag_to_ix[START_TAG]  # Sanity check
        assert start == self.START_TAG_IDX  # Sanity check
        best_path.reverse()
        return path_score, best_path

    #  def neg_log_likelihood(self, sentence, tags):
    #      feats = self._get_lstm_features(sentence)
    def neg_log_likelihood(self, feats, tags):
        forward_score = self._forward_alg(feats)
        gold_score = self._score_sentence(feats, tags)
        return forward_score - gold_score

    #  def forward(self, sentence):  # dont confuse this with _forward_alg above.
    #      # Get the emission scores from the BiLSTM
    #      feats = self._get_lstm_features(sentence)

    def forward(self, feats):

        # Find the best path, given the features.
        score, tag_seq = self._viterbi_decode(feats)
        return score, tag_seq


if __name__ == '__main__':
    START_TAG = "<START>"
    STOP_TAG = "<STOP>"
    EMBEDDING_DIM = 300
    HIDDEN_DIM = 256

    # Make up some training data
    training_data = [(
        "the wall street journal reported today that apple corporation made money"
        .split(), "B I I I O O O B I O O".split()),
                     ("georgia tech is a university in georgia".split(),
                      "B I O O O O B".split())]

    word_to_ix = {}
    for sentence, tags in training_data:
        for word in sentence:
            if word not in word_to_ix:
                word_to_ix[word] = len(word_to_ix)

    tag_to_ix = {"B": 0, "I": 1, "O": 2, START_TAG: 3, STOP_TAG: 4}

    model = BiLSTM_CRF(len(word_to_ix), tag_to_ix, EMBEDDING_DIM, HIDDEN_DIM)
    optimizer = optim.SGD(model.parameters(), lr=0.01, weight_decay=1e-4)

    # Check predictions before training
    with torch.no_grad():
        precheck_sent = prepare_sequence(training_data[0][0], word_to_ix)
        precheck_tags = torch.tensor(
            [tag_to_ix[t] for t in training_data[0][1]], dtype=torch.long)
        print(model(precheck_sent))

    # Make sure prepare_sequence from earlier in the LSTM section is loaded
    for epoch in range(
            300
    ):  # again, normally you would NOT do 300 epochs, it is toy data
        for sentence, tags in training_data:
            # Step 1. Remember that Pytorch accumulates gradients.
            # We need to clear them out before each instance
            model.zero_grad()

            # Step 2. Get our inputs ready for the network, that is,
            # turn them into Tensors of word indices.
            sentence_in = prepare_sequence(sentence, word_to_ix)
            targets = torch.tensor([tag_to_ix[t] for t in tags],
                                   dtype=torch.long)

            # Step 3. Run our forward pass.
            loss = model.neg_log_likelihood(sentence_in, targets)

            # Step 4. Compute the loss, gradients, and update the parameters by
            # calling optimizer.step()
            loss.backward()
            optimizer.step()

    # Check predictions after training
    with torch.no_grad():
        precheck_sent = prepare_sequence(training_data[0][0], word_to_ix)
        print(model(precheck_sent))
        # We got it!



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from loguru import logger
from collections import Counter

import torch
import torch.nn as nn
from torch.nn import CrossEntropyLoss
import torch.nn.functional as F

from ...losses import FocalLoss, LabelSmoothingCrossEntropy

from .utils import CNerTokenizer, SeqEntityScore, get_entities
from ..models.linears import PoolerStartLogits, PoolerEndLogits
from ..trainer import Trainer, get_default_optimizer_parameters
from ...utils.multiprocesses import barrier_leader_process, barrier_member_processes, is_multi_processes

from torch.utils.data import DataLoader, RandomSampler, SequentialSampler
from transformers import AutoConfig, BertConfig, BertTokenizer, BertTokenizerFast, AutoModelForTokenClassification
from transformers.modeling_bert import BertPreTrainedModel, BertModel


class BertSpanForNer(BertPreTrainedModel):
    def __init__(self, config):
        super(BertSpanForNer, self).__init__(config)
        self.soft_label = config.soft_label
        self.num_labels = config.num_labels
        self.loss_type = config.loss_type
        self.focalloss_gamma = config.focalloss_gamma
        self.focalloss_alpha = config.focalloss_alpha

        self.bert = BertModel(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.start_fc = PoolerStartLogits(config.hidden_size, self.num_labels)
        if self.soft_label:
            self.end_fc = PoolerEndLogits(config.hidden_size + self.num_labels,
                                          self.num_labels)
        else:
            self.end_fc = PoolerEndLogits(config.hidden_size + 1,
                                          self.num_labels)
        self.init_weights()

    def forward(self,
                input_ids,
                attention_mask=None,
                token_type_ids=None,
                start_positions=None,
                end_positions=None):
        #  subjects_ids=None):
        outputs = self.bert(input_ids=input_ids,
                            attention_mask=attention_mask,
                            token_type_ids=token_type_ids)
        sequence_output = outputs[0]
        sequence_output = self.dropout(sequence_output)
        start_logits = self.start_fc(sequence_output)
        if start_positions is not None and self.training:
            if self.soft_label:
                batch_size = input_ids.size(0)
                seq_len = input_ids.size(1)
                label_logits = torch.FloatTensor(batch_size, seq_len,
                                                 self.num_labels)
                label_logits.zero_()
                label_logits = label_logits.to(input_ids.device)
                label_logits.scatter_(2, start_positions.unsqueeze(2), 1)
            else:
                label_logits = start_positions.unsqueeze(2).float()
        else:
            label_logits = F.softmax(start_logits, -1)
            if not self.soft_label:
                label_logits = torch.argmax(label_logits,
                                            -1).unsqueeze(2).float()
        end_logits = self.end_fc(sequence_output, label_logits)
        avg_logits = (start_logits + end_logits) / 2
        outputs = (
            #  avg_logits,
            start_logits,
            end_logits,
        ) + outputs[2:]

        if start_positions is not None and end_positions is not None:
            assert self.loss_type in [
                'LabelSmoothingCrossEntropy', 'FocalLoss', 'CrossEntropyLoss'
            ]
            if self.loss_type == 'LabelSmoothingCrossEntropy':
                loss_fct = LabelSmoothingCrossEntropy()
            elif self.loss_type == 'FocalLoss':
                loss_fct = FocalLoss(gamma=self.focalloss_gamma,
                                     alpha=self.focalloss_alpha)
            else:
                loss_fct = CrossEntropyLoss()
            start_logits = start_logits.view(-1, self.num_labels)
            end_logits = end_logits.view(-1, self.num_labels)
            active_loss = attention_mask.view(-1) == 1
            active_start_logits = start_logits[active_loss]
            active_end_logits = end_logits[active_loss]

            active_start_labels = start_positions.view(-1)[active_loss]
            active_end_labels = end_positions.view(-1)[active_loss]

            start_loss = loss_fct(active_start_logits, active_start_labels)
            end_loss = loss_fct(active_end_logits, active_end_labels)
            total_loss = (start_loss + end_loss) / 2
            outputs = (total_loss, ) + outputs
            return outputs
        else:
            #  return (0.0, ) + outputs
            #  return (torch.tensor(0.0).cuda(), ) + outputs
            return outputs


class SpanEntityScore(object):
    def __init__(self, id2label):
        self.id2label = id2label
        self.reset()

    def reset(self):
        self.origins = []
        self.founds = []
        self.rights = []

    def compute(self, origin, found, right):
        recall = 0 if origin == 0 else (right / origin)
        precision = 0 if found == 0 else (right / found)
        f1 = 0. if recall + precision == 0 else (2 * precision *
                                                 recall) / (precision + recall)
        return recall, precision, f1

    def result(self):
        class_info = {}
        origin_counter = Counter([self.id2label[x[0]] for x in self.origins])
        found_counter = Counter([self.id2label[x[0]] for x in self.founds])
        right_counter = Counter([self.id2label[x[0]] for x in self.rights])
        for type_, count in origin_counter.items():
            origin = count
            found = found_counter.get(type_, 0)
            right = right_counter.get(type_, 0)
            recall, precision, f1 = self.compute(origin, found, right)
            class_info[type_] = {
                "acc": round(precision, 4),
                'recall': round(recall, 4),
                'f1': round(f1, 4)
            }
        origin = len(self.origins)
        found = len(self.founds)
        right = len(self.rights)
        recall, precision, f1 = self.compute(origin, found, right)
        return {'acc': precision, 'recall': recall, 'f1': f1}, class_info

    def update(self, true_subject, pred_subject):
        self.origins.extend(true_subject)
        self.founds.extend(pred_subject)
        self.rights.extend([
            pre_entity for pre_entity in pred_subject
            if pre_entity in true_subject
        ])


MODEL_CLASSES = {
    'bert': (BertConfig, BertSpanForNer, CNerTokenizer),
    #  'bert': (BertConfig, BertSpanForNer, BertTokenizer),
    #  'bert': (BertConfig, BertSpanForNer, BertTokenizerFast),
}


def load_pretrained_tokenizer(args):
    config_class, model_class, tokenizer_class = MODEL_CLASSES[args.model_type]
    tokenizer = tokenizer_class.from_pretrained(
        args.model_path,
        do_lower_case=args.do_lower_case,
        cache_dir=args.cache_dir if args.cache_dir else None,
    )

    return tokenizer


def load_pretrained_model(args):
    # make sure only the first process in distributed training
    # will download model & vocab
    barrier_member_processes(args)

    config_class, model_class, tokenizer_class = MODEL_CLASSES[args.model_type]
    config = config_class.from_pretrained(
        args.model_path,
        num_labels=args.num_labels,
        loss_type=args.loss_type,
        cache_dir=args.cache_dir if args.cache_dir else None,
        soft_label=args.soft_label,
    )
    setattr(config, 'soft_label', args.soft_label)
    setattr(config, 'loss_type', args.loss_type)
    setattr(config, 'focalloss_gamma', args.focalloss_gamma)
    setattr(config, 'focalloss_alpha', args.focalloss_alpha)
    logger.info(f"model_path: {args.model_path}")
    logger.info(f"config:{config}")
    model = model_class.from_pretrained(
        args.model_path,
        from_tf=bool(".ckpt" in args.model_path),
        config=config,
        cache_dir=args.cache_dir if args.cache_dir else None,
    )

    # make sure only the first process in distributed training
    # will download model & vocab
    barrier_leader_process(args)

    return model


def collate_fn(batch):
    """
    batch should be a list of (sequence, target, length) tuples...
    Returns a padded tensor of sequences sorted from longest to shortest,
    """
    all_input_ids, all_input_mask, all_segment_ids, all_start_ids, all_end_ids, all_lens = map(
        torch.stack, zip(*batch))
    max_len = max(all_lens).item()
    all_input_ids = all_input_ids[:, :max_len]
    all_input_mask = all_input_mask[:, :max_len]
    all_segment_ids = all_segment_ids[:, :max_len]
    all_start_ids = all_start_ids[:, :max_len]
    all_end_ids = all_end_ids[:, :max_len]
    return all_input_ids, all_input_mask, all_segment_ids, all_start_ids, all_end_ids, all_lens


def bert_extract_item(start_logits, end_logits):
    S = []
    start_pred = torch.argmax(start_logits, -1).cpu().numpy()[0][1:-1]
    end_pred = torch.argmax(end_logits, -1).cpu().numpy()[0][1:-1]
    for i, s_l in enumerate(start_pred):
        if s_l == 0:
            continue
        for j, e_l in enumerate(end_pred[i:]):
            if s_l == e_l:
                S.append((s_l, i, i + j))
                break
    return S


def load_model(args):
    model = load_pretrained_model(args)
    model.to(args.device)
    return model


def build_default_model(args):
    """
    自定义模型
    规格要求返回模型(model)、优化器(optimizer)、调度器(scheduler)三元组。
    """

    # -------- model --------
    model = load_pretrained_model(args)
    model.to(args.device)

    # -------- optimizer --------
    from transformers.optimization import AdamW
    optimizer_parameters = get_default_optimizer_parameters(
        model, args.weight_decay)
    optimizer = AdamW(optimizer_parameters,
                      lr=args.learning_rate,
                      correct_bias=False)

    # -------- scheduler --------
    from transformers.optimization import get_linear_schedule_with_warmup

    scheduler = get_linear_schedule_with_warmup(
        optimizer,
        num_warmup_steps=args.total_steps * args.warmup_rate,
        num_training_steps=args.total_steps)

    return model, optimizer, scheduler


def init_labels(args, labels):

    args.ner_labels = ['[unused1]'] + labels
    args.id2label = {i: label for i, label in enumerate(args.ner_labels)}
    args.label2id = {label: i for i, label in enumerate(args.ner_labels)}
    args.num_labels = len(args.label2id)

    logger.info(f"args.label2id: {args.label2id}")
    logger.info(f"args.id2label: {args.id2label}")
    logger.info(f"args.num_labels: {args.num_labels}")


class NerTrainer(Trainer):
    def __init__(self, args, ner_labels, build_model=None, tokenizer=None):
        super(NerTrainer, self).__init__(args)
        init_labels(args, ner_labels)
        if tokenizer:
            self.tokenizer = tokenizer
        else:
            self.tokenizer = load_pretrained_tokenizer(args)

        if build_model is None:
            self.build_model = build_default_model
        else:
            self.build_model = build_model

        self.label2id = args.label2id
        self.collate_fn = collate_fn

    def examples_to_dataset(self, examples, max_seq_length):
        from .dataset import examples_to_dataset
        return examples_to_dataset(examples, self.label2id, self.tokenizer,
                                   max_seq_length)

    def batch_to_inputs(self, args, batch, known_labels=True):
        inputs = {
            "input_ids": batch[0],
            "attention_mask": batch[1],
            "start_positions": batch[3],
            "end_positions": batch[4],
        }
        if args.model_type != "distilbert":
            # XLM and RoBERTa don"t use segment_ids
            inputs["token_type_ids"] = (
                batch[2] if args.model_type in ["bert", "xlnet"] else None)

        return inputs

    #  def generate_dataloader(self, args, dataset, batch_size, keep_order=True):
    #
    #      Sampler = SequentialSampler if keep_order else RandomSampler
    #      sampler = DistributedSampler(dataset) if is_multi_processes(
    #          args) else Sampler(dataset)
    #      dataloader = DataLoader(dataset,
    #                              sampler=sampler,
    #                              batch_size=batch_size,
    #                              collate_fn=collate_fn)
    #      return dataloader

    def on_eval_start(self, args, eval_dataset):
        self.metric = SpanEntityScore(args.id2label)

    #  def on_eval_step(self, args, eval_dataset, step, model, inputs, outputs):
    def on_eval_step(self, args, model, step, batch, batch_features):
        all_input_ids, all_input_mask, all_segment_ids, all_start_ids, all_end_ids, all_lens = batch

        eval_loss = 0.0
        num_eval_steps = 0
        for i in range(all_input_ids.size()[0]):
            inputs = {
                "input_ids": all_input_ids[i].view(1, -1),
                "attention_mask": all_input_mask[i].view(1, -1),
                "start_positions": all_start_ids[i].view(1, -1),
                "end_positions": all_end_ids[i].view(1, -1),
            }
            if args.model_type != "distilbert":
                # XLM and RoBERTa don"t use segment_ids
                inputs["token_type_ids"] = (all_segment_ids[i].view(
                    1, -1) if args.model_type in ["bert", "xlnet"] else None)

            outputs = model(**inputs)
            #  tmp_eval_loss, avg_logits, start_logits, end_logits = outputs[:4]
            tmp_eval_loss, start_logits, end_logits = outputs[:3]
            eval_loss += tmp_eval_loss
            num_eval_steps += 1

            R = bert_extract_item(start_logits, end_logits)

            T = batch_features[i].subjects
            self.metric.update(true_subject=T, pred_subject=R)

        eval_loss = eval_loss / num_eval_steps
        eval_info, entity_info = self.metric.result()
        results = {f'{key}': value for key, value in eval_info.items()}
        results['loss'] = eval_loss

        return (eval_loss, ), results

    def on_predict_start(self, args, test_dataset):
        self.pred_results = []

    def on_predict_step(self, args, model, step, batch):
        #  inputs = self.batch_to_inputs(args, batch, known_labels=False)
        #  outputs = model(**inputs)
        #  _, start_logits, end_logits = outputs[:3]
        #  R = bert_extract_item(start_logits, end_logits)
        all_input_ids, all_input_mask, all_segment_ids, all_start_ids, all_end_ids, all_lens = batch

        for i in range(all_input_ids.size()[0]):
            inputs = {
                "input_ids": all_input_ids[i].view(1, -1),
                "attention_mask": all_input_mask[i].view(1, -1),
                "start_positions": all_start_ids[i].view(1, -1),
                "end_positions": all_end_ids[i].view(1, -1),
            }
            if args.model_type != "distilbert":
                # XLM and RoBERTa don"t use segment_ids
                inputs["token_type_ids"] = (all_segment_ids[i].view(
                    1, -1) if args.model_type in ["bert", "xlnet"] else None)

            outputs = model(**inputs)
            #  _, _, start_logits, end_logits = outputs[:4]
            _, start_logits, end_logits = outputs[:3]

            R = bert_extract_item(start_logits, end_logits)

            if R:
                label_entities = [[args.id2label[x[0]], x[1], x[2]] for x in R]
            else:
                label_entities = []

            #  logger.debug(f"{label_entities}")
            json_d = {}
            json_d['id'] = step
            #  tag_seq = [args.id2label[x] for x in preds]
            #  json_d['tag_seq'] = " ".join(tag_seq)
            json_d['entities'] = label_entities

            #  logger.debug(f"{json_d}")

            self.pred_results.append(json_d)

    def on_predict_end(self, args, test_dataset):
        return self.pred_results

    def on_eval_end(self, args, eval_dataset):
        from ...utils.ner_utils import get_ner_results
        results = get_ner_results(self.metric)
        return results



#!/usr/bin/env python
# -*- coding: utf-8 -*-


class Config:
    SEED = 8864
    debug_extract_labels = False

    surfix = ""

    experiment = "resumes"
    data_dir = f"../user_data/seqlabel/{experiment}"
    #  results_dir = f"../prediction_result/{experiment}_result/seqlabel"
    output_dir = f"../user_data/{experiment}_output/seqlabel"

    # -------------------- seqlabel model train & test begin. ---------------
    #  #  category = "WORKCONTENT"
    #  #  top_k = 20
    #  #  s_threshold = 0.1
    #  #  e_threshold = 0.01
    #  #  min_label_len = 8
    #  #  seg_len = 510
    #  #  seg_backoff = 400
    #  #  strictly_equal = False
    #  #  use_additional_section = True
    #  #  num_augments = 0
    #  #  surfix = f"_aug{num_augments}"
    #
    #  category = "PROJRESP"
    #  top_k = 20
    #  s_threshold = 0.1
    #  e_threshold = 0.1  #0.1
    #  min_label_len = 8
    #  seg_len = 510
    #  seg_backoff = 400
    #  use_additional_section = True
    #  strictly_equal = False
    #  num_augments = 0
    #  surfix = f"_aug{num_augments}"
    #
    #  #  category = "PNAME"
    #  #  top_k = 20
    #  #  s_threshold = 0.1
    #  #  e_threshold = 0.1
    #  #  min_label_len = 8
    #  #  seg_len = 510
    #  #  seg_backoff = 64
    #  #  use_additional_section = False
    #  #  strictly_equal = False
    #  #  num_augments = 0
    #  #  surfix = f"_aug{num_augments}"
    #
    #  #  category = "WORKUNIT"
    #  #  top_k = 20
    #  #  s_threshold = 0.1
    #  #  e_threshold = 0.01
    #  #  min_label_len = 8
    #  #  seg_len = 300
    #  #  seg_backoff = 150
    #  #  use_additional_section = False
    #  #  strictly_equal = False
    #  #  num_augments = 0
    #  #  surfix = f"_aug{num_augments}"
    #
    #  #  category = "WORKDUTY"
    #  #  top_k = 20
    #  #  s_threshold = 0.1
    #  #  e_threshold = 0.01
    #  #  min_label_len = 8
    #  #  seg_len = 510
    #  #  seg_backoff = 64
    #  #  use_additional_section = False
    #  #  strictly_equal = False
    #  #  num_augments = 0
    #  #  surfix = f"_aug{num_augments}"
    #
    #  raw_train_data_file = f"{data_dir}/train_{category}{surfix}.tsv"
    #  raw_test_data_file = f"{data_dir}/test_{category}{surfix}.tsv"
    #  #  raw_test_data_file = "test_one.tsv"
    #
    #  total_folds = 5
    #  train_folds_file = f"{output_dir}/train_{category}_{total_folds}folds{surfix}.pkl"
    #
    #  # 是否应该取训练数据最大长度或最大长度95%（99%）？
    #  max_seq_length = seg_len  # 510  #160

    # -------------------- seqlabel model train & test end. ---------------

    workcontent_model = f"../user_data/resumes_output/seqlabel/best_model_WORKCONTENT_fold0.weights"
    projresp_model = f"../user_data/resumes_output/seqlabel/best_model_PROJRESP_fold0.weights"

    #  predict_model_dir = f"../user_data/outputs/resumes_output/bert/checkpoint-5476"
    #  ner_model_0_dir = f"../user_data/outputs/resumes_output/bert/checkpoint-5476"
    #  ner_model_1_dir = f"../user_data/outputs/resumes_output/bert/checkpoint-4107"
    #  ner_model_2_dir = f"../user_data/outputs/resumes_output/bert/checkpoint-6845"
    #  predict_model_dir = f"../user_data/outputs/resumes_output/bert/checkpoint-6840"
    #  ner_model_0_dir = f"../user_data/outputs/resumes_output/bert/checkpoint-6840"
    #  ner_model_1_dir = f"../user_data/outputs/resumes_output/bert/checkpoint-5130"
    #  ner_model_2_dir = f"../user_data/outputs/resumes_output/bert/checkpoint-8550"
    predict_model_dir = f"predict_model_dir_auto_defined_in_predict_py"
    ner_model_0_dir = f"ner_model_0_dir_auto_defined_in_predict_py"
    ner_model_1_dir = f"ner_model_1_dir_auto_defined_in_predict_py"
    ner_model_2_dir = f"ner_model_2_dir_auto_defined_in_predict_py"

    BERT_CHINESE_DIR = "../data/External/pretrained/tensorflow/bert-base-chinese"

    epochs = 6  # 训练轮次
    batch_size = 8  # 按GPU显存大小调整，建议8G 取16, 12G 取24，16G 取32

    learning_rate = 5e-5
    min_learning_rate = 1e-5
    #  learning_rate = 1e-5
    #  min_learning_rate = 1e-6

    # 反序文本和标签，训练新模型
    reverse_text = False
    # ls -l /dev/nvidia? | wc -l
    gpus = 1

    # TPU
    steps_per_update = 8  # 权重更新前梯度累积的步数。
    #  learning_rate = 5e-4  # 运用梯度累积时，适当调高学习率

    config_path = f"{BERT_CHINESE_DIR}/bert_config.json"
    checkpoint_path = f"{BERT_CHINESE_DIR}/bert_model.ckpt"
    dict_path = f"{BERT_CHINESE_DIR}/vocab.txt"


def get_best_model_file_name(selected_fold: int) -> str:
    return f"{Config.output_dir}/best_model_fold{selected_fold}{surfix}.model"


def get_best_model_weights_file_name(selected_fold: int, category) -> str:
    return f"{Config.output_dir}/best_model_{category}_fold{selected_fold}{Config.surfix}.weights"


#  def get_result_file_name(selected_fold: int) -> str:
#      return f"{Config.results_dir}/result_{Config.category}_fold{selected_fold}{Config.surfix}.txt"


def get_wrong_eval_file_name(selected_fold: int) -> str:
    return f"{Config.output_dir}/wrong_eval_fold{selected_fold}{Config.surfix}.txt"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from loguru import logger
from ..common_args import get_main_args


def add_modeling_args(parser):
    return parser


def get_args(special_args: list = None):
    return get_main_args(add_modeling_args, special_args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from loguru import logger
from ..common_args import get_main_args


def add_modeling_args(parser):
    parser.add_argument('--markup',
                        default='bios',
                        type=str,
                        choices=['bios', 'bio'])
    parser.add_argument("--autofix",
                        action="store_true",
                        help="Auto fix CRF label errors.")
    parser.add_argument(
        "--bios_file",
        type=str,
        default=None,
        help="The bios file path.",
    )
    parser.add_argument(
        "--crf_type",
        type=str,
        default='ncrfpp',
        help=
        "CRF type. ['ncrfpp', 'pytorch-crf', 'old_crf', 'new_crf', 'lstm_crf']"
    )
    parser.add_argument("--no_crf_loss",
                        action="store_true",
                        help="Do not calculate CRF loss.")

    return parser


def get_args(special_args: list = None):
    return get_main_args(add_modeling_args, special_args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json, copy
from pathlib import Path
from tqdm import tqdm
from loguru import logger
import numpy as np
import torch
from torch.utils.data import DataLoader, TensorDataset
from torch.utils.data import RandomSampler, SequentialSampler
from ...utils.multiprocesses import barrier_leader_process, barrier_member_processes, is_multi_processes
from ..ner_utils import InputExample


class InputFeatures(object):
    """A single set of features of data."""
    def __init__(self, input_ids, input_mask, input_len, segment_ids,
                 label_ids):
        self.input_ids = input_ids
        self.input_mask = input_mask
        self.segment_ids = segment_ids
        self.label_ids = label_ids
        self.input_len = input_len

    def __repr__(self):
        return str(self.to_json_string())

    def to_dict(self):
        """Serializes this instance to a Python dictionary."""
        output = copy.deepcopy(self.__dict__)
        return output

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        #  return json.dumps(self.to_dict(), indent=2, sort_keys=True) + "\n"
        return json.dumps(self.to_dict(), ensure_ascii=False,
                          sort_keys=True) + "\n"



def to_BIOS(labels, num_tokens):
    bios_labels = ['O'] * num_tokens
    #  logger.debug(f"num_tokens: {num_tokens}, labels: {labels}")
    for c, s, e in labels:
        assert s <= e
        if s == e:
            bios_labels[s] = f"S-{c}"
        else:
            bios_labels[s] = f"B-{c}"
            for i in range(1, e + 1 - s):
                bios_labels[s + i] = f"I-{c}"

    return bios_labels


def encode_examples(examples, label2id, tokenizer, max_seq_length):

    texts = [e.text_a[:max_seq_length - 2] for e in examples]

    #  for e in tqdm(examples, desc="to BIOS"):
    #      e.labels = to_BIOS(e.labels, len(e.text_a))

    #  logger.debug(f"texts: {texts}")

    #  outputs = tokenizer.batch_encode_plus(texts,
    #                                        max_length=max_seq_length,
    #                                        add_special_tokens=True,
    #                                        return_tensors='pt',
    #                                        pad_to_max_length=True,
    #                                        return_attention_mask=True,
    #                                        return_token_type_ids=True)
    #  all_input_ids, all_attention_mask, all_token_type_ids = outputs[
    #      'input_ids'], outputs['attention_mask'], outputs['token_type_ids']

    all_tokens = [[tokenizer.cls_token] + tokenizer.tokenize(text) +
                  [tokenizer.sep_token]
                  for text in tqdm(texts, desc="Tokenize")]

    #  all_text_lens = [len(text) for text in texts]
    #  logger.debug(f"all_text_lens: {all_text_lens}")

    all_input_lens = [len(tokens) for tokens in all_tokens]
    #  logger.debug(f"all_input_lens: {all_input_lens}")

    all_input_ids = [
        tokenizer.convert_tokens_to_ids(tokens) for tokens in all_tokens
    ]
    all_attention_mask = [[1] * len(input_ids) for input_ids in all_input_ids]
    all_token_type_ids = [[0] * len(input_ids) for input_ids in all_input_ids]

    def encode_labels(labels, input_len):
        if labels is not None:
            #  logger.debug(f"labels: {labels}")
            label_ids = [label2id[x] for x in to_BIOS(labels, input_len)]
        else:
            label_ids = [label2id['O']] * input_len
        return label_ids

    all_labels = [[label2id[tokenizer.cls_token]] +
                  encode_labels(e.labels, input_len - 2) +
                  [label2id[tokenizer.sep_token]]
                  for e, input_len in zip(examples, all_input_lens)]

    all_padding_lens = [max_seq_length - n for n in all_input_lens]
    for i, (input_ids, attention_mask, token_type_ids, label_ids,
            padding_length) in enumerate(
                zip(all_input_ids, all_attention_mask, all_token_type_ids,
                    all_labels, all_padding_lens)):
        all_input_ids[i] = input_ids + [0] * padding_length
        all_attention_mask[i] = attention_mask + [0] * padding_length
        all_token_type_ids[i] = token_type_ids + [0] * padding_length
        all_labels[i] = label_ids + [0] * padding_length

    logger.debug(f"all_input_ids.shape: {np.array(all_input_ids).shape}")
    logger.debug(
        f"all_attention_mask.shape: {np.array(all_attention_mask).shape}")
    logger.debug(
        f"all_token_type_ids.shape: {np.array(all_token_type_ids).shape}")
    logger.debug(f"all_labels.shape: {np.array(all_labels).shape}")
    logger.debug(f"all_input_lens.shape: {np.array(all_input_lens).shape}")
    assert np.array(all_input_ids).shape[1] == max_seq_length
    assert np.array(all_attention_mask).shape[1] == max_seq_length
    assert np.array(all_token_type_ids).shape[1] == max_seq_length

    all_features = [
        InputFeatures(input_ids=input_ids,
                      input_mask=attention_mask,
                      segment_ids=token_type_ids,
                      label_ids=label_ids,
                      input_len=input_len)
        for input_ids, attention_mask, token_type_ids, label_ids, input_len in
        zip(all_input_ids, all_attention_mask, all_token_type_ids, all_labels,
            all_input_lens)
    ]

    return (all_features, {
        'input_ids':
        torch.tensor(all_input_ids, dtype=torch.long),
        'attention_mask':
        torch.tensor(all_attention_mask, dtype=torch.long),
        'token_type_ids':
        torch.tensor(all_token_type_ids, dtype=torch.long),
        'labels':
        torch.tensor(all_labels, dtype=torch.long),
        'input_lens':
        torch.tensor(all_input_lens, dtype=torch.long),
    })


def examples_to_dataset(examples, label2id, tokenizer, max_seq_length):
    all_features, outputs = encode_examples(examples, label2id, tokenizer,
                                            max_seq_length)

    dataset = TensorDataset(
        outputs['input_ids'],
        outputs['attention_mask'],
        outputs['token_type_ids'],
        outputs['labels'],
        outputs['input_lens'],
    )

    return dataset, all_features


#  def convert_examples_to_features(
#      examples,
#      label2id,
#      tokenizer,
#      max_seq_length=512,
#      cls_token_at_end=False,
#      cls_token="[CLS]",
#      cls_token_segment_id=1,
#      sep_token="[SEP]",
#      pad_on_left=False,
#      pad_token=0,
#      pad_token_segment_id=0,
#      sequence_a_segment_id=0,
#      mask_padding_with_zero=True,
#  ):
#      """ Loads a data file into a list of `InputBatch`s
#          `cls_token_at_end` define the location of the CLS token:
#              - False (Default, BERT/XLM pattern): [CLS] + A + [SEP] + B + [SEP]
#              - True (XLNet/GPT pattern): A + [SEP] + B + [SEP] + [CLS]
#          `cls_token_segment_id` define the segment id associated to the CLS token (0 for BERT, 2 for XLNet)
#      """
#      #  label2id = {label: i for i, label in enumerate(examples.label_set)}
#      features = []
#      for ex_index, e in enumerate(tqdm(examples, desc="Convert examples")):
#          tokens = tokenizer.tokenize(e.text_a)
#          if e.labels is not None:
#              label_ids = [label2id[x] for x in e.labels]
#          else:
#              label_ids = [label2id['O']] * len(e.text_a)
#
#          # Account for [CLS] and [SEP] with "- 2".
#          special_tokens_count = 2
#          if len(tokens) > max_seq_length - special_tokens_count:
#              tokens = tokens[:(max_seq_length - special_tokens_count)]
#              label_ids = label_ids[:(max_seq_length - special_tokens_count)]
#
#          # The convention in BERT is:
#          # (a) For sequence pairs:
#          #  tokens:   [CLS] is this jack ##son ##ville ? [SEP] no it is not . [SEP]
#          #  type_ids:   0   0  0    0    0     0       0   0   1  1  1  1   1   1
#          # (b) For single sequences:
#          #  tokens:   [CLS] the dog is hairy . [SEP]
#          #  type_ids:   0   0   0   0  0     0   0
#          #
#          # Where "type_ids" are used to indicate whether this is the first
#          # sequence or the second sequence. The embedding vectors for `type=0` and
#          # `type=1` were learned during pre-training and are added to the wordpiece
#          # embedding vector (and position vector). This is not *strictly* necessary
#          # since the [SEP] token unambiguously separates the sequences, but it makes
#          # it easier for the model to learn the concept of sequences.
#          #
#          # For classification tasks, the first vector (corresponding to [CLS]) is
#          # used as as the "sentence vector". Note that this only makes sense because
#          # the entire model is fine-tuned.
#          tokens += [sep_token]
#          label_ids += [label2id[sep_token]]
#          segment_ids = [sequence_a_segment_id] * len(tokens)
#
#          if cls_token_at_end:
#              tokens += [cls_token]
#              label_ids += [label2id[cls_token]]
#              segment_ids += [cls_token_segment_id]
#          else:
#              tokens = [cls_token] + tokens
#              label_ids = [label2id[cls_token]] + label_ids
#              segment_ids = [cls_token_segment_id] + segment_ids
#
#          input_ids = tokenizer.convert_tokens_to_ids(tokens)
#          # The mask has 1 for real tokens and 0 for padding tokens. Only real
#          # tokens are attended to.
#          input_mask = [1 if mask_padding_with_zero else 0] * len(input_ids)
#          input_len = len(label_ids)
#          # Zero-pad up to the sequence length.
#          padding_length = max_seq_length - len(input_ids)
#          if pad_on_left:
#              input_ids = ([pad_token] * padding_length) + input_ids
#              input_mask = ([0 if mask_padding_with_zero else 1] *
#                            padding_length) + input_mask
#              segment_ids = ([pad_token_segment_id] *
#                             padding_length) + segment_ids
#              label_ids = ([pad_token] * padding_length) + label_ids
#          else:
#              input_ids += [pad_token] * padding_length
#              input_mask += [0 if mask_padding_with_zero else 1] * padding_length
#              segment_ids += [pad_token_segment_id] * padding_length
#              label_ids += [pad_token] * padding_length
#
#          assert len(input_ids) == max_seq_length
#          assert len(input_mask) == max_seq_length
#          assert len(segment_ids) == max_seq_length
#          assert len(label_ids) == max_seq_length
#          if ex_index < 3:
#              logger.debug("*** Example ***")
#              logger.debug(f"guid: {e.guid}")
#              logger.debug(f"tokens: {' '.join([str(x) for x in tokens])}")
#              logger.debug(f"input_ids: {' '.join([str(x) for x in input_ids])}")
#              logger.debug(
#                  f"input_mask: {' '.join([str(x) for x in input_mask])}")
#              logger.debug(
#                  f"segment_ids: {' '.join([str(x) for x in segment_ids])}")
#              logger.debug(f"label_ids: {' '.join([str(x) for x in label_ids])}")
#
#          features.append(
#              InputFeatures(input_ids=input_ids,
#                            input_mask=input_mask,
#                            input_len=input_len,
#                            segment_ids=segment_ids,
#                            label_ids=label_ids))
#      return features
#
#
#  def features_to_dataset(features):
#      # Convert to Tensors and build dataset
#      all_input_ids = torch.tensor([f.input_ids for f in features],
#                                   dtype=torch.long)
#      all_input_mask = torch.tensor([f.input_mask for f in features],
#                                    dtype=torch.long)
#      all_segment_ids = torch.tensor([f.segment_ids for f in features],
#                                     dtype=torch.long)
#      all_label_ids = torch.tensor([f.label_ids for f in features],
#                                   dtype=torch.long)
#      all_lens = torch.tensor([f.input_len for f in features], dtype=torch.long)
#      dataset = TensorDataset(all_input_ids, all_input_mask, all_segment_ids,
#                              all_lens, all_label_ids)
#      return dataset
#
#
#  def examples_to_dataset_old(args,
#                              examples,
#                              ner_labels,
#                              tokenizer,
#                              max_seq_length,
#                              evaluate=False,
#                              alias=""):
#      # Make sure only the first process in distributed training process the dataset,
#      # and the others will use the cache
#      if not evaluate:
#          barrier_member_processes(args)
#
#      label_set = ['X', 'O', '[CLS]', '[SEP]']
#      for x in ner_labels:
#          label_set += [f"B-{x}", f"I-{x}", f"S-{x}"]
#
#      args.id2label = {i: label for i, label in enumerate(label_set)}
#      args.label2id = {label: i for i, label in enumerate(label_set)}
#      args.num_labels = len(label_set)
#
#      # Load data features from cache or dataset file
#      cached_features_file = Path(
#          args.data_dir
#      ) / f"cached_ner_{args.dataset_name}{'_' + alias if alias else ''}"
#      if args.cache_features and cached_features_file.exists(
#      ) and not args.overwrite_cache:
#          logger.info(
#              f"Loading features from cached file {cached_features_file}")
#          features = torch.load(cached_features_file)
#      else:
#          logger.info(f"Creating features from dataset file at {args.data_dir}")
#
#          features = convert_examples_to_features(
#              examples,
#              args.label2id,
#              tokenizer,
#              max_seq_length=max_seq_length,
#              cls_token_at_end=bool(args.model_type in ["xlnet"]),
#              pad_on_left=bool(args.model_type in ['xlnet']),
#              cls_token=tokenizer.cls_token,
#              cls_token_segment_id=2 if args.model_type in ["xlnet"] else 0,
#              sep_token=tokenizer.sep_token,
#              # pad on the left for xlnet
#              pad_token=tokenizer.convert_tokens_to_ids([tokenizer.pad_token
#                                                         ])[0],
#              pad_token_segment_id=4 if args.model_type in ['xlnet'] else 0,
#          )
#
#          if args.cache_features and is_master_process(args):
#              logger.info(
#                  f"Saving features into cached file {cached_features_file}")
#              torch.save(features, cached_features_file)
#
#      # Make sure only the first process in distributed training process the dataset,
#      # and the others will use the cache
#      if not evaluate:
#          barrier_leader_process(args)
#
#      dataset = features_to_dataset(features)
#
#      return dataset


def load_examples_from_bios_file(filename):
    def _read_txt(input_file):
        lines = []
        with open(input_file, 'r') as f:
            words = []
            labels = []
            for line in f:
                if line.startswith("-DOCSTART-") or line == "" or line == "\n":
                    if words:
                        lines.append({"words": words, "labels": labels})
                        words = []
                        labels = []
                else:
                    splits = line.split(" ")
                    words.append(splits[0])
                    if len(splits) > 1:
                        labels.append(splits[-1].replace("\n", ""))
                    else:
                        # Examples could have no label for mode = "test"
                        labels.append("O")
            if words:
                lines.append({"words": words, "labels": labels})
        return lines

    def _create_examples(lines):
        """Creates examples for the training and dev sets."""
        examples = []
        for (i, line) in enumerate(lines):
            #  if i == 0:
            #      continue
            guid = "%s-%s" % ("example", i)
            text_a = line['words']
            # BIOS
            labels = []
            for x in line['labels']:
                if 'M-' in x:
                    labels.append(x.replace('M-', 'I-'))
                elif 'E-' in x:
                    labels.append(x.replace('E-', 'I-'))
                else:
                    labels.append(x)
            examples.append(
                InputExample(guid=guid, text_a=text_a, labels=labels))
        return examples

    return _create_examples(_read_txt(filename))


def export_bios_file(examples, bios_file):
    fw_bios = open(bios_file, 'w')
    for example in tqdm(examples):
        # for test examples
        labels = example.labels
        if labels is None:
            labels = ['O'] * len(example.text_a)

        for w, t in zip(example.text_a, labels):
            fw_bios.write(f"{w} {t}\n")
        fw_bios.write("\n")
    logger.info(f"export {len(examples)} examples to {bios_file}")


# ------ Generate CRF train data -------

[
    {
        'id': "",
        'text': "",
        'entities': [
            {
                'label': "",
                'start_pos': "",
                'end_pos': "",
            },
        ]
    },
]


class TaggedCRFData:
    def __init__(self, all_text_entities):
        self.all_text_entities = all_text_entities
        self.labels_map = []

    def generate_tagged_crf_data(self, num_augements=10, shuffle=False):
        lines = self._generate_crf_data()
        aug_lines = self._augement(num_augements, shuffle=shuffle)
        return lines + aug_lines

    def generate_tagged_text(self, text, entities):
        num_tokens = len(text)
        tokens = ['O'] * num_tokens
        p0 = 0
        for entity in entities:
            label_type = entity['label_type']
            s = entity['start_pos'] - 1
            e = entity['end_pos'] - 1
            print(s, e)
            assert e >= s
            if s >= len(text) or e >= len(text):
                continue

            entity_text = text[s:e + 1]
            print(label_type, entity_text)

            assert len(entity_text) > 0
            if label_type not in self.labels_map:
                self.labels_map[label_type] = []
            self.labels_map[label_type].append(entity_text)

            n = e - s + 1
            t = label2type[label_type]
            if n == 1:
                tokens[s] = f"S-{t}"
            else:
                tokens[s] = f"B-{t}"
                for i in range(1, n):
                    tokens[s + i] = f"I-{t}"
        return tokens

    def _generate(self):
        self.labels_map = {}
        lines = []
        for json_file, text, entities in self.all_text_entities:
            tokens = generate_tagged_text(text, entities, self.labels_map)
            chars = [c for c in text]
            for n, (c, tok) in enumerate(zip(chars, tokens)):
                #  Fout.write(f"{n+1:03d}: {c} {tok}\n")
                lines.append(f"{c} {tok}\n")
            lines.append("\n")
        return lines

    def _augement(self, num_augements=10, shuffle=False):
        aug_tokens = []
        for i, (json_file, text, entities) in enumerate(
                tqdm(self.all_text_entities,
                     desc=f"Augement {num_augements}X")):

            #  print(f"-------------------{json_file}--------------------")
            #  print(text)
            #  print(entities)
            #  for entity in entities:
            #      s = entity['start_pos'] - 1
            #      e = entity['end_pos']
            #      print(f"{entity['label_type']}: {text[s:e]}")
            #  print("----------------------------------------")
            for _ in range(num_augements):
                e_idx = random.randint(0, len(entities) - 1)
                entity = entities[e_idx]

                label_type = entity['label_type']
                s = entity['start_pos'] - 1
                e = entity['end_pos'] - 1

                labels = self.labels_map[label_type]
                idx = random.randint(0, len(labels) - 1)
                new_entity_text = labels[idx]

                text = text[:s] + new_entity_text + text[e:]
                #  print(entity, new_entity_text)
                assert len(new_entity_text) >= 0
                delta = len(new_entity_text) - (e - s + 1) + 1

                #  print(
                #      f"delta={delta} e_idx: {e_idx} new_entity_text: {new_entity_text}"
                #  )
                #  print(entities)
                entity['end_pos'] = entity['start_pos'] + len(
                    new_entity_text) - 1
                for n, e in enumerate(entities):
                    if n > e_idx:
                        e['start_pos'] += delta
                        e['end_pos'] += delta
                        assert e['start_pos'] <= e['end_pos']
                    else:
                        assert e['start_pos'] <= e['end_pos']
                #  print(f"-- Adjusted {entities}")
                #  print(text[:s] + "【" + new_entity_text + "】" +
                #        text[entity['end_pos'] + 1:])

                tokens = generate_tagged_text(text, entities)
                aug_tokens.append((text, tokens))

        if shuffle:
            random.shuffle(aug_tokens)

        lines = []

        for i, (text,
                tokens) in enumerate(tqdm(aug_tokens, desc='Write augements')):
            chars = [c for c in text]
            for n, (c, tok) in enumerate(zip(chars, tokens)):
                lines.append(f"{c} {tok}\n")
            lines.append("\n")

        return lines



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import torch
import numpy as np
from loguru import logger
import onnxruntime

from .trainer import generate_dataloader
from ..utils.progbar import Progbar


def export_onnx(args,
                model,
                dataset,
                onnx_file,
                input_names=None,
                output_names=None):
    inputs = tuple([x.cuda() for x in list(dataset.tensors)[:-1]])
    torch.onnx.export(model,
                      args=inputs,
                      f=onnx_file,
                      input_names=input_names,
                      output_names=output_names)

    logger.info(f"Export onnx model to {onnx_file}.")


def inference_from_onnx(args, onnx_file, test_dataset):
    test_dataloader = generate_dataloader(args, test_dataset,
                                          args.per_gpu_predict_batch_size)
    session = onnxruntime.InferenceSession(onnx_file, None)
    inputs = session.get_inputs()
    outputs = session.get_outputs()
    for i, input in enumerate(inputs):
        logger.debug(f"onnx inputs[{i}]: {input}")
    for i, output in enumerate(outputs):
        logger.debug(f"onnx outputs[{i}]: {output}")

    pbar = Progbar(target=len(test_dataloader), desc=f"Predicting")

    logits = []
    for step, batch in enumerate(test_dataloader):
        batch = tuple(t.to(args.device) for t in batch)
        all_input_ids, all_attention_mask, all_token_type_ids, all_labels = batch

        onnx_args = {
            inputs[i].name: x.detach().cpu().numpy()
            for i, x in enumerate(batch[:-1])
        }

        batch_logits = [session.run([], onnx_args)[0].tolist()[0]]

        logits += batch_logits
        pbar.update(step + 1)

    logits = np.array(logits)
    logger.debug(f"logits[:10]: {logits[:10]}")

    return logits



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable

#  class FocalLoss(nn.Module):
#
#      # gamma: focusing parameter
#      # alpha: balance parameter
#      def __init__(self, gamma=2, alpha=0.25):
#          super(FocalLoss, self).__init__()
#
#          self.gamma = gamma
#          self.alpha = alpha
#
#      def forward(self, output, target):
#
#          cross_entropy = F.cross_entropy(output, target)
#          cross_entropy_log = torch.log(cross_entropy)
#          logpt = - F.cross_entropy(output, target)
#          pt    = torch.exp(logpt)
#
#          focal_loss = -((1 - pt) ** self.gamma) * logpt
#
#          focal_loss = self.alpha * focal_loss
#
#          return focal_loss


class FocalLoss(nn.Module):
    def __init__(self, gamma=2.0, alpha=None, size_average=True):
        super(FocalLoss, self).__init__()
        self.gamma = gamma
        self.alpha = alpha
        #  if isinstance(alpha,(float,int,long)): self.alpha = torch.Tensor([alpha,1-alpha])
        if isinstance(alpha, (float, int)):
            self.alpha = torch.Tensor([alpha, 1 - alpha])
        if isinstance(alpha, list): self.alpha = torch.Tensor(alpha)
        self.size_average = size_average

    def forward(self, input, target):
        #  print(input)
        #  print(target)
        #  print(f"input.shape: {input.shape}, target.shape: {target.shape}")
        if input.dim() > 2:
            input = input.view(input.size(0), input.size(1),
                               -1)  # N,C,H,W => N,C,H*W
            input = input.transpose(1, 2)  # N,C,H*W => N,H*W,C
            input = input.contiguous().view(
                -1, input.size(2))  # N,H*W,C => N*H*W,C
        target = target.view(-1, 1)

        logpt = F.log_softmax(input)
        logpt = logpt.gather(1, target)
        logpt = logpt.view(-1)
        pt = Variable(logpt.data.exp())

        if self.alpha is not None:
            if self.alpha.type() != input.data.type():
                self.alpha = self.alpha.type_as(input.data)
            at = self.alpha.gather(0, target.data.view(-1))
            logpt = logpt * Variable(at)

        loss = -1 * (1 - pt)**self.gamma * logpt
        if self.size_average: return loss.mean()
        else: return loss.sum()


#  class FocalLoss(nn.Module):
#      '''Multi-class Focal loss implementation'''
#      def __init__(self, gamma=2, alpha=None, ignore_index=-100):
#          super(FocalLoss, self).__init__()
#          self.gamma = gamma
#          self.alpha = alpha
#          self.ignore_index = ignore_index
#
#      def forward(self, input, target):
#          """
#          input: [N, C]
#          target: [N, ]
#          """
#          logpt = F.log_softmax(input, dim=1)
#          pt = torch.exp(logpt)
#          logpt = (1 - pt)**self.gamma * logpt
#          loss = F.nll_loss(logpt,
#                            target,
#                            self.alpha,
#                            ignore_index=self.ignore_index)
#          return loss
